
# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest, sys
from timeit import default_timer as timer
from test_expert_interface import MicroSpecTestExpertInterface
from microspeclib.expert  import MicroSpecExpertInterface

@pytest.mark.skipif(sys.platform not in ["darwin","linux"], reason="Emulation currently only runs on linux and MacOS")
class MicroSpecTestExpertInterfaceEmulator(MicroSpecTestExpertInterface):
  __test__ = True

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.software = MicroSpecExpertInterface(emulation=True, timeout=1)















# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

__doc__ = """Automated and static docstrings for json-generated classes, functions, and globals

Most of the API objects in MicroSpecLib are dynamically generated from cfg/microspec.json.
For example CommandSetBridgeLED, MicroSpecSimpleInterface.setBridgeLED, and LEDGreen are not
defined in code, they are created by meta-class factories, so that if the configuration changes
to permit new protocol commands and replies, the functions and classes will automatically update
as well. However, this causes a problem for documentation. 

Great pains were taken to auto-generate the proper function and class signatures such as
CommandSetBridgeLED(led_num=None, led_setting=None) rather than just CommandSetBridgeLED(\*args,
\**kwargs), so that pydoc and sphinx could introspect and document them properly. However,
the ONE thing that cannot be auto-generated is the per-function and per-class human-read
documentation. It could go into yet another json file and be auto-generated, but that would
be no more compact than making a file elsewhere for it. It could be written directly in
sphinx doc/source/\*.rst files, but then pydoc wouldn't find them. So the only choice left
is to make a separate internals docstrings library with the data in one place there, and
that is what this module is. Luckily, it at least saves some repetition, since both
CommandSetBridgeLED and MicroSpecSimpleInterface.setBridgeLED need the same documentation,
for example.

Thus, this module contains dictionaries for the different data and class types, and the datatype
module imports and implements them when it instantiates the metaclasses and generates the
globals. Since the json protocol and the documentation are separate, we must assume that one might
get out of sync with the other, thus it is assumed that this documentation may be missing
something. As such, implementers of these dictionaries should use DOC.get(value, None) and
handle lack of documentation in a responsible way.

The CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetBridgeLED"] contains that docstring, for example.

And the _common global is used to hold common replacement patterns, to eliminate having to type
the same things over and over below - they are replaced at the end of the module.

"""

CHROMASPEC_DYNAMIC_DOC = {"command":{}, "bridge":{}, "sensor":{}}

_common = { 
  "dt": "microspeclib.datatypes", 
  "led_status": """led_status: :data:`~{dt}.types.LEDOff`, :data:`~{dt}.types.LEDGreen`, or :data:`~{dt}.types.LEDRed`""" + \
                """  The color state of the LED""",
  "status":     """status: :data:`~{dt}.types.StatusOK` or :data:`~{dt}.types.StatusError`""" + \
                """  If there is an error status, the other attributes are not valid""",
  "useful":     """the useful API return values""",
  "notfinal":   """This is not the final payload for this command type. If the Simple or Expert API returns this object, """ + \
                """then the command failed in the Bridge and did not even make it to the Sensor."""
}

CHROMASPEC_DYNAMIC_DOC["command"]["CommandNull"] = """Null loopback request. The hardware will not reply,
but the API will still return a null reply object. This is primarily used to flush the line
in case of desynchronization.

Returns
-------
:class:`~{dt}.bridge.BridgeNull`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetBridgeLED"] = """Retrieves the current state of an LED on
the Bridge board.

Parameters
----------
led_num: 0
  The index of the LED on the Bridge

Returns
-------
:class:`~{dt}.bridge.BridgeGetBridgeLED`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandSetBridgeLED"] = """Sets the current state of an LED on the Bridge board.

Parameters
----------
led_num: 0
  The index of the LED on the Bridge
{led_status}

Returns
-------
:class:`~{dt}.bridge.BridgeSetBridgeLED`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetSensorLED"] = """Retrieves the current state of an LED on
the Sensor board.

Parameters
----------
led_num: 0-2
  The index of the LED on the Sensor

Returns
-------
:class:`~{dt}.bridge.BridgeGetSensorLED`
:class:`~{dt}.bridge.SensorGetSensorLED`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandSetSensorLED"] = """Sets the current state of an LED on the Bridge board.

Parameters
----------
led_num: 0-2
  The index of the LED on the Sensor
{led_status}

Returns
-------
:class:`~{dt}.bridge.BridgeSetSensorLED`
:class:`~{dt}.bridge.SensorSetSensorLED`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandReset"] = """Resets the hardware and replies when the reset is complete.

Returns
-------
:class:`~{dt}.bridge.BridgeReset`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandVerify"] = """Verifies running status of the hardware.

Returns
-------
:class:`~{dt}.bridge.BridgeVerify`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetSensorConfig"] = """Retrieves the current sensor configuration.

Returns
-------
:class:`~{dt}.bridge.BridgeGetSensorConfig`
:class:`~{dt}.sensor.SensorGetSensorConfig`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandSetSensorConfig"] = """Sets the current sensor configuration.

Parameters
----------
binning: 0-1
  Whether or not to bin related pixel values
gain: 0-255
  Gain
row_bitmap:
  Which rows to permit sensing on. There are 5, and can all be activated with a binary bitmap of 5 1's, i.e. 011111 or 0x1F.
  Any combination is permitted except 0x0

Returns
-------
:class:`~{dt}.bridge.BridgeSetSensorConfig`
:class:`~{dt}.sensor.SensorSetSensorConfig`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandAutoExposure"] = """
Autoexpose the spectrometer: find the exposure time that results in the peak
signal in the configured pixel range to land in the configured target signal
range.

The autoexpose algorithm runs in the dev-kit firmware. Configure autoexpose
parameters with command `setAutoExposeConfig()`, see
:class:`~{dt}.command.CommandSetAutoExposeConfig`.

Returns success (0 or 1) and number of exposure times attempted.

Does not return the final exposure time. Get exposure time with getExposure(),
see :class:`~{dt}.command.CommandGetExposure`.

Example
-------
>>> from microspeclib.simple import MicroSpecSimpleInterface
>>> kit = MicroSpecSimpleInterface(timeout=2.0)
>>> reply = kit.autoExposure()
>>> print(reply.success) # 1 means success
1
>>> print(reply.iterations) # number of exposure times before success/give-up
3

Returns
-------
:class:`~{dt}.bridge.BridgeAutoExposure`
:class:`~{dt}.sensor.SensorAutoExposure` ← {useful}

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetAutoExposeConfig"] = """Retrieves the current auto-expose configuration.

Example
-------
>>> from microspeclib.simple import MicroSpecSimpleInterface
>>> kit = MicroSpecSimpleInterface(timeout=2.0)
>>> reply = kit.getAutoExposeConfig()
>>> print(reply.max_tries)
>>> print(reply.start_pixel)
>>> print(reply.stop_pixel)
>>> print(reply.target)
>>> print(reply.target_tolerance)

Returns
-------
:class:`~{dt}.bridge.BridgeGetAutoExposeConfig`
:class:`~{dt}.sensor.SensorGetAutoExposeConfig`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandSetAutoExposeConfig"] = """Sets the current auto-expose configuration.

Parameters
----------
max_tries: 1-255
  Maximum number of exposures to try before giving up.

  Firmware defaults to 10 on power-up.

  If max_tries is 0, status is ERROR and the AutoExposeConfig is
  not changed.

start_pixel: 7-392 if binning on, 14-784 if binning off
  Auto-expose ignores pixels below start_pixel when checking if
  the peak is in the target range.

  Firmware defaults to 7 on power-up.
  Recommended value is the smallest pixel number in the
  pixel-to-wavelength map.

  If start_pixel is outside the allowed range, status is ERROR
  and the AutoExposeConfig is not changed.

stop_pixel: 7-392 if binning on, 14-784 if binning off, must be >= start_pixel
  Auto-expose ignores pixels above stop_pixel when checking if
  the peak is in the target range.

  Firmware defaults to 392 on power-up.
  Recommended value is the largest pixel number in the
  pixel-to-wavelength map.

  If stop_pixel < start_pixel, or if stop_pixel is outside the
  allowed range, status is ERROR and the AutoExposeConfig is not
  changed.

target: 4500-65535
  Target peak-counts for exposure gain calculation.

  Firmware defaults to 46420 counts on power-up.

  If target is outside the allowed range, status is ERROR and the
  AutoExposeConfig is not changed.

target_tolerance: 0-65535
  target +/- target_tolerance is the target peak-counts range.
  Auto-expose is finished when the peak counts lands in this
  range.

  Firmware defaults to 3277 counts on power-up.

  If the combination of target and target_tolerance results in
  target ranges extending below 4500 counts or above 65535
  counts, auto-expose ignores the target_tolerance and clamps the
  target range at these boundaries.

Returns
-------
:class:`~{dt}.bridge.BridgeSetAutoExposeConfig`
:class:`~{dt}.sensor.SensorSetAutoExposeConfig`

"""

CHROMASPEC_DYNAMIC_DOC["command"]["CommandGetExposure"] = """Retrieve the current exposure setting, which may have been set
either by :class:`~{dt}.command.CommandSetExposure` or :class:`~{dt}.command.CommandAutoExposure`.

Returns
-------
:class:`~{dt}.bridge.BridgeGetExposure`
:class:`~{dt}.sensor.SensorGetExposure`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandSetExposure"] = """Set the exposure value for the sensor.

Parameters
----------
cycles: 1-65535
  Number of cycles to wait to collect pixel strength.

Returns
-------
:class:`~{dt}.bridge.BridgeSetExposure`
:class:`~{dt}.sensor.SensorSetExposure`

"""
CHROMASPEC_DYNAMIC_DOC["command"]["CommandCaptureFrame"] = """Retrieve one set of captured pixels.

Returns
-------
:class:`~{dt}.bridge.BridgeCaptureFrame`
:class:`~{dt}.sensor.SensorCaptureFrame`

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeNull"] = """This packet doesn't actually exist, as the request for a Null
has no reply. However, to distinguish between an error, when a :class:`~{dt}.command.CommandNull` is requested, the
API returns this object rather than None.

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeGetBridgeLED"] = """Contains the result of a :class:`~{dt}.command.CommandGetBridgeLED`
command.

Parameters
----------
{status}
led_num: 0
  Which LED the status applies to
{led_status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeSetBridgeLED"] = """Contains the status of the :class:`~{dt}.command.CommandSetBridgeLED`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeGetSensorLED"] = """Contains a transitory status of the :class:`~{dt}.command.CommandGetSensorLED`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeSetSensorLED"] = """Contains a transitory status of the :class:`~{dt}.command.CommandGetSensorLED`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeReset"] = """Contains status status of the :class:`~{dt}.command.CommandReset`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeVerify"] = """Contains the status of the :class:`~{dt}.command.CommandVerify`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeGetSensorConfig"] = """Contains a transitory status of the :class:`~{dt}.command.CommandGetSensorConfig`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeSetSensorConfig"] = """Contains a transitory status of the :class:`~{dt}.command.CommandSetSensorConfig`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeAutoExposure"] = """Contains a transitory status of the :class:`~{dt}.command.CommandAutoExposure`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeGetExposure"] = """Contains a transitory status of the :class:`~{dt}.command.CommandGetExposure`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeSetExposure"] = """Contains a transitory status of the :class:`~{dt}.command.CommandSetExposure`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["bridge"]["BridgeCaptureFrame"] = """Contains a transitory status of the :class:`~{dt}.command.CommandCaptureFrame`
command as it passes through the Bridge. {notfinal}

Parameters
----------
{status}

"""

CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorGetSensorLED"] = """Contains the result of a :class:`~{dt}.command.CommandGetSensorLED`
command.

Parameters
----------
{status}
led_num: 0 or 1
  Which LED the status applies to
{led_status}

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorSetSensorLED"] = """Contains the status of the :class:`~{dt}.command.CommandSetSensorLED`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorGetSensorConfig"] = """Contains the result of a :class:`~{dt}.command.CommandGetSensorConfig`
command.

Parameters
----------
{status}
binning: 0-1
  Whether or not to bin adjacent pixels.
  0: binning off, LIS-770i has 784 7.8µm-pitch pixels, 770 optically active
  1: binning on, LIS-770i has 392 15.6µm-pitch pixels, 385 optically active
gain: 0-255
  Analog pixel voltage gain. Allowed values:
  0x01: 1x gain
  0x25: 2.5x gain (37 in decimal)
  0x04: 4x gain
  0x05: 5x gain
row_bitmap:
  Which rows to permit sensing on. There are 5, and can all be
  activated with a binary bitmap of 5 1's, i.e. 011111 or 0x1F.
  The three most significant bits must be 0. Otherwise, any
  combination is permitted except 0x00.

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorSetSensorConfig"] = """Contains the result of a :class:`~{dt}.command.CommandSetSensorConfig`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorAutoExposure"] = """Contains the result of a :class:`~{dt}.command.CommandAutoExposure`
command.

Parameters
----------
{status}

success: 0 or 1

  1: success
    Auto-expose settled on an exposure time that put the sensor
    peak value within the target range.

  0: failure
    Auto-expose gave up on finding an exposure time, either
    because it reached the maximum number of tries or the
    exposure time is already at the maximum allowed value.

iterations: 1-255

  Number of exposures tried by auto-expose.

  `iterations` never exceeds AutoExposeConfig parameter
  `max_tries`, as this sets the maximum number of iterations to
  try.

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorGetExposure"] = """Contains the result of a :class:`~{dt}.command.CommandGetExposure`
command.

Parameters
----------
{status}
cycles: 1-65535
  Number of cycles to expose pixels. Each cycle is 20µs.

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorSetExposure"] = """Contains the result of a :class:`~{dt}.command.CommandSetExposure`
command.

Parameters
----------
{status}

"""
CHROMASPEC_DYNAMIC_DOC["sensor"]["SensorCaptureFrame"] = """Contains the result of a :class:`~{dt}.command.CommandCaptureFrame`
command.

Parameters
----------
{status}
num_pixels: 0-784
  The number of pixels to expect in the pixels parameter.
  Using the recommended (default) configuration, num_pixels is
  392.
pixels: Array of values, each from 0-65535
  The pixel values for one capture frame.

"""

for protocol in CHROMASPEC_DYNAMIC_DOC:
  for klass in CHROMASPEC_DYNAMIC_DOC[protocol]:
    while [found for found in _common.keys() if "{%s}"%found in CHROMASPEC_DYNAMIC_DOC[protocol][klass]]:
      #import pdb; pdb.set_trace()
      CHROMASPEC_DYNAMIC_DOC[protocol][klass] = CHROMASPEC_DYNAMIC_DOC[protocol][klass].format(**_common)



import tkinter as tk
from microspeclib.datatypes.command import *
from microspeclib.datatypes.types   import *
from microspeclib.expert            import MicroSpecExpertInterface as expert

class Spectrometer(tk.Canvas):
  def __init__(self, *args, **kwargs):
    tk.Canvas.__init__(self, *args, **kwargs)
    #self.create_rectangle(50, 50, 100, 100)

  def redraw(self, pixels):
    if not pixels:
      return
    self.update() # or else the width and height commands won't work
    self.width  = self.winfo_width()
    self.height = self.winfo_height()
    npixels = len(pixels)
    minpixel = None
    maxpixel = None
    for p in pixels:
      minpixel = p if minpixel is None or p < minpixel else minpixel
      maxpixel = p if maxpixel is None or p > maxpixel else maxpixel
    if minpixel is None or maxpixel is None:
      return
    pixelrange = maxpixel - minpixel
    self.delete("all")
    n = 0
    for p in pixels:
      yrelative = p - minpixel
      ynormalized = yrelative * self.height / pixelrange
      xnormalized = n * self.width / npixels
      self.draw_dot(xnormalized, ynormalized)
      n = n + 1

  def draw_dot(self, x, y):
      self.create_rectangle(x,   self.height - y, 
                            x+1, self.height - y+1)

class BridgeLED(tk.Frame):
  def __init__(self, *args, **kwargs):
    tk.Frame.__init__(self, *args, **kwargs)
    self.state = tk.IntVar()
    self.label = tk.Label(self, text="Bridge LED", justify=tk.LEFT, padx=10)
    self.off   = tk.Radiobutton(self, text="Off",   padx=10, variable=self.state, value=LEDOff,  command=lambda:self.usb_send(self.state.get()))
    self.green = tk.Radiobutton(self, text="Green", padx=10, variable=self.state, value=LEDGreen,command=lambda:self.usb_send(self.state.get()))
    self.red   = tk.Radiobutton(self, text="Red",   padx=10, variable=self.state, value=LEDRed,  command=lambda:self.usb_send(self.state.get()))
    self.label.grid(row=0, column=0)
    self.off.grid(  row=0, column=1)
    self.green.grid(row=0, column=2)
    self.red.grid(  row=0, column=3)

  def usb_send(self, state):
    self.master.push_command(CommandSetBridgeLED(led_num=0, led_setting=state))

  def usb_receive(self, reply):
    if reply and reply.status == 0 and hasattr(reply, "led_setting"):
      self.set(reply.led_setting)

  def set(self, state):
    self.state.set(state)

  def get(self):
    return self.state.get()
  
class MicrospecGUI(tk.Frame):
  def __init__(self, *args, emulation=False, **kwargs):
    tk.Frame.__init__(self, *args, **kwargs)
    self.spec = Spectrometer(self, width=400, height=600)
    self.led  = BridgeLED(self)

    self.spec.grid(row=0, column=0)
    self.led.grid( row=0, column=1)

    self.queue = []
    self.wait  = []

    self.usb  = expert(emulation=emulation)

    self.push_command(CommandGetBridgeLED(led_num=0), self.led)
    self.push_command(CommandSetExposure(cycles=100))
    self.push_command(CommandSetSensorConfig(binning=True, gain=100, row_bitmap=0x1F))

    self.main_thread()

  def push_command(self, command, requestor=None):
    self.queue.append((command, requestor))

  def usb_receive(self, reply):
    if reply and reply.status == 0:
      #print("pixels: %s"%(reply))
      self.spec.redraw(reply.pixels)
      pass

  def main_thread(self):
    self.push_command(CommandCaptureFrame(), self)

    reply = self.usb.receiveReply()
    while reply:
      recipient = self.wait.pop(0)
      if recipient[1]:
        recipient[1].usb_receive(reply)
      reply = self.usb.receiveReply()

    while self.queue:
      command = self.queue.pop(0)
      self.usb.sendCommand(command[0])
      self.wait.append(command)

    self.after(100, self.main_thread)
    
    
    

root = tk.Tk()
gui = MicrospecGUI(root, emulation=False)
gui.pack()
gui.mainloop()



import pytest

# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

@pytest.fixture(scope="class")
def class_results(request):
  request.cls.results = [["Command", "Time(ms)"]]





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.simple            import MicroSpecSimpleInterface
from microspeclib.logger            import debug, CHROMASPEC_LOGGER as log
from microspeclib.datatypes.command import CHROMASPEC_COMMAND_NAME as commands
from microspeclib.datatypes.types   import CHROMASPEC_GLOBAL       as constants
import time, datetime, sys

# NOTE: There is no docstring because the Sphinx documentation runs the argparse help and creates the
#       documentation there. See the Sphinx documentation for the microspec_cmdline.py executable.

cmdname = dict([ [s[7:].lower(), s[7:]] for s    in commands.keys() ])
cstname = dict([ [s    .lower(), v    ] for s, v in constants.items() ])

def get_command(c):
  return cmdname.get(c.lower())

def get_constant(c):
  return cstname.get(c.lower(), c)

def print_format(csv, response):
  current_time = datetime.datetime.now().strftime("%Y-%m-%dT%T.%f%z")
  if response is None:
    print("%s,%s"%(current_time,None))
  elif csv:
    print("%s,%s"%(current_time,response.csv()))
  else:
    print("%s,%s"%(current_time,str(response)))

def command_help_string():
  cmdstring = "List of commands and arguments:\n"
  for cname, cmd in commands.items():
    cmdstring += "  %-17s %s\n"%(cname[7:], " ".join(["%s=xxx"%(k) for k in cmd().variables if k != "command_id"]))
  cmdstring += "\nNote that command names are case-insensitive. For more information on each command, run `pydoc microspeclib.datatypes.command`"
  return cmdstring

if __name__ == "__main__":
  import argparse
  parser = argparse.ArgumentParser(
    description="Command line interface for MicroSpecLib", 
    epilog=command_help_string(),
    formatter_class=argparse.RawDescriptionHelpFormatter )
  parser.add_argument("-d", "--debug",    help="Internal debugging trace",           action="count", default=0    )
  parser.add_argument("-v", "--verbose",  help="Verbose trace",                      action="count", default=0    )
  parser.add_argument("-t", "--timeout",  help="Timeout (seconds)",                  type=float,     default=0.1  )
  parser.add_argument("-r", "--repeat",   help="Repeat N times, 1=once, 0=forever",  type=int,       default=1    )
  parser.add_argument("-w", "--wait",     help="Wait inbetween repeats (seconds)",   type=float,     default=1.0  )
  parser.add_argument("-e", "--emulator", help="Spawn emulator and connect to that", action="count", default=0    )
  parser.add_argument("-f", "--file",     help="File/socket/device to connect to, "+
                                               "default=auto-detect hardware",                       default=None )
  parser.add_argument("-c", "--csv",      help="Print-format: 'default' or 'csv'",   action="count", default=0    )
  parser.add_argument("command",          help="Command to send",                                                 )
  parser.add_argument("arguments",        help="Key=value pairs for command",        nargs="*",      default=[]   )
  import sys
  if "-i" in sys.argv or "--ignore" in sys.argv:
    parser.add_argument("-i", "--ignore", help="Ignore argument") # hidden argument for internal use
  args = parser.parse_args()
  
  from microspeclib.internal.stream import MicroSpecEmulatedStream
  from microspeclib.logger          import debug
  if args.debug or args.verbose: 
    debug(args.debug>0)

  timeout = args.timeout

  si = None
  if args.emulator:
    log.info("Starting emulation process...")
    hardware = MicroSpecEmulatedStream(socat=True, fork=True, timeout=0.1)
    software = MicroSpecSimpleInterface(device=hardware.software, timeout=timeout)
    log.info("Connecting to emulation")
    si       = software
  elif args.file:
    log.info("Connecting to device '%s'", args.file)
    si       = MicroSpecSimpleInterface(device=args.file, timeout=timeout)
  else:
    log.info("Connecting to default hardware")
    si       = MicroSpecSimpleInterface(timeout=timeout)

  try:
    getcmd  = get_command(args.command)
    command = getcmd[0].lower() + getcmd[1:]
  except:
    log.critical("Command '%s' not found!\n%s", args.command, command_help_string())
    sys.exit(1)

  try:
    kwargs  = dict([ [s.split("=")[0],get_constant(s.split("=")[1])]  for s in args.arguments ])
  except:
    log.critical("Error parsing arguments '%s'", str(args.arguments))
    sys.exit(1)

  i = 0
  while (True if args.repeat < 0 else i < args.repeat):
    log.info("Executing command %s(%s)", command, kwargs)
    try:
      reply = getattr(si, command)(**kwargs)
      print_format(args.csv, reply)
    except Exception as e:
      log.error(e)
      print_format(args.csv, None)
    i += 1
    if (True if args.repeat < 0 else i < args.repeat):
      if args.repeat > 0:
        log.info("%d calls remain", args.repeat - i + 1)
      log.info("Waiting %f seconds...", args.wait)
      time.sleep(args.wait)




#!/usr/bin/env python

# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

if __name__ == "__main__":
  from microspeclib.simple    import MicroSpecSimpleInterface
  from microspeclib.datatypes import *
  
  si = MicroSpecSimpleInterface(timeout=0.1)
  #si = MicroSpecSimpleInterface(timeout=0.1, device="COM3")
  #si = MicroSpecSimpleInterface(timeout=0.1, device="/dev/cu.usbserial-CHROMATION09310")
  #si = MicroSpecSimpleInterface(timeout=0.1, emulation=True)
  
  print(si.setBridgeLED(led_num=0, led_setting=LEDOff))
  print(si.setSensorLED(led_num=0, led_setting=LEDGreen))
  print(si.setSensorLED(led_num=1, led_setting=LEDRed))
  print(si.getBridgeLED(led_num=0))
  print(si.getSensorLED(led_num=0))
  reply = si.getSensorLED(led_num=1)
  print(reply.status)
  print(reply.led_setting)
  
  print(si.setSensorConfig(binning=True, gain=Gain1x, row_bitmap=0x1F))
  print(si.setExposure(cycles=100))
  
  import time
  for i in range(0,5):
    reply = si.captureFrame()
    print(reply.status)
    print(reply.num_pixels)
    pixels = [str(pixel) for pixel in reply.pixels]
    print("first 3 pixels: %s"%(",".join(pixels[:3])))
    time.sleep(0.5)
  



#!/usr/bin/env python

"""
Example Usage
=============

In the examples below, Windows users replace ".py" with ".exe".

Capture one frame once
----------------------
microspec_cmdline.py captureframe

Capture a frame 50 times with 2 seconds inbetween
-------------------------------------------------
microspec_cmdline.py captureframe -r 50 -w 2

Capture a frame and print results in csv
----------------------------------------
microspec_cmdline.py captureframe -c

Set the binning to true, gain to 100, and row_bitmap to 010101 (0x15)
---------------------------------------------------------------------
microspec_cmdline.py setsensorconfig binning=1 gain=100 row_bitmap=0x15

Connect to a specific COM4 port
-------------------------------
microspec_cmdline.py -f COM4 ...

Connect to a specific /dev/com123 file
--------------------------------------
microspec_cmdline.py -f /dev/com123 ...

Connect to emulator instead of hardware
---------------------------------------
microspec_cmdline.py -e ...

"""

def main():
  import subprocess, sys
  # subprocess.call(["python", "-m", "microspeclib.cmdline"] + sys.argv[1:])
  # Use sys.executable instead of "python" to see the
  # microspeclib package installed in the active virtual
  # environment.
  subprocess.call([sys.executable, "-m", "microspeclib.cmdline"] + sys.argv[1:])

if __name__ == "__main__":
  main()





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.datatypes import CommandNull
from microspeclib.logger    import CHROMASPEC_LOGGER as log
import time

# The intended difference between this and the Simple interface is to provide more
# fine control, such as breaking up sending commands and then looping to wait for
# replies. It requires creating Command objects and then passing them along, in 
# contrast to the one routine per command structure of the Simple interface.

__all__ = [ "MicroSpecExpertInterface" ]

from microspeclib.internal.stream import MicroSpecSerialIOStream, MicroSpecEmulatedStream
class MicroSpecExpertInterface(MicroSpecSerialIOStream):
  def __init__(self, serial_number=None, device=None, timeout=2.0, retry_timeout=0.001, emulation=False, *args, **kwargs):
    log.info("serial_number=%s, device=%s, timeout=%s, retry_timeout=%s, emulation=%s, args=%s, kwargs=%s",
             serial_number, device, timeout, retry_timeout, emulation, args, kwargs)
    if emulation:
      self.emulation = MicroSpecEmulatedStream(socat=True, fork=True, timeout=timeout)
      device = self.emulation.software
    super().__init__(serial_number=serial_number, 
                     device=device, timeout=timeout, 
                     *args, **kwargs)
    self.retry_timeout = retry_timeout
    self.current_command = []
    log.info("return")

  def __setattr__(self, attr, value):
    if attr == "timeout":
      log.info("set timeout to %s", value)
      self.stream.timeout = value
    else:
      self.__dict__[attr] = value
      
  def __getattribute__(self, attr):
    if attr == "timeout":
      log.info("return timeout=%s", self.stream.timeout)
      return self.stream.timeout
    else:
      return super().__getattribute__(attr)
      
  def sendCommand(self, command):
    """Send one command to the sensor, do not wait for a response. You may stack these, and they
    will return in FIFO order later.

    Parameters
    ----------
    command: :py:mod:`CommandObject <microspeclib.datatypes.command>`
      Create one of these objects and pass it as an argument

    Returns
    -------
    None

    """
    log.info("command=%s", command)
    try:
      if bytes(command) == b'':
        log.warning("Error packing payload for command '%s'", str(command))
        raise Exception("Unable to send partial command '%s'"%(str(command)))
    except:
      log.warning("Error packing payload for command '%s'", str(command))
      raise Exception("Unable to send partial command '%s'"%(str(command)))
    super().sendCommand(command)
    log.info("appending command=%s to current_command=%s", command, self.current_command)
    self.current_command.append(command)
    log.info("return")

  def receiveReply(self):
    """Receive one command result command sent. If multiple were sent, receive the next one in FIFO order,
    if available. A failure to retrieve it does not remove it from the queue, but a reply that contains
    a failed status does.

    Returns
    -------
    :py:mod:`BridgeReplyObject <microspeclib.datatypes.bridge>`
    :py:mod:`SensorReplyObject <microspeclib.datatypes.sensor>`

    """
    log.info("waiting for reply")
    start = time.time()
    reply = super().receiveReply(self.current_command[0].command_id) if self.current_command else None
    since = time.time() - start
    timeout = self.timeout if self.timeout else 0
    remain = timeout - since
    log.info("start=%s reply=%s since=%s timeout=%s remain=%s"%(start,reply,since,timeout,remain))
    while reply is None and remain > 0:
      log.info("no reply yet, timeout remaining=%s", remain)
      time.sleep( self.retry_timeout if remain > self.retry_timeout else remain )
      reply = super().receiveReply(self.current_command[0].command_id) if self.current_command else None
      since = time.time() - start
      remain = timeout - since
      log.info("start=%s reply=%s since=%s timeout=%s remain=%s"%(start,reply,since,timeout,remain))
    if reply:
      log.info("popping command since reply was found")
      self.current_command.pop(0)
    log.info("return %s", reply)
    return reply

  def sendAndReceive(self, command):
    """Send one command to the sensor, and receive one command result command sent. 
    If multiple were sent, receive the next one in FIFO order, if available. 
    This does NOT mean that the reply will match the command, if you already sent a command
    that you had not received yet. A failure to retrieve it does not remove it from
    the queue, but a reply that contains a failed status does.

    Parameters
    ----------
    command: :py:mod:`CommandObject <microspeclib.datatypes.command>`
      Create one of these objects and pass it as an argument

    Returns
    -------
    :py:mod:`BridgeReplyObject <microspeclib.datatypes.bridge>`
    :py:mod:`SensorReplyObject <microspeclib.datatypes.sensor>`

    """
    log.info("command=%s", command)
    self.sendCommand(command)
    reply = self.receiveReply()
    log.info("return %s", reply)
    return reply

  def flush(self, timeout=0.1):
    """Tell hardware to finish sending any remaining commands and flush it's communication line,
    then flush the local queue and throw away all waiting replies. This does not send a Reset command."""
    log.info("flushing stream by sending a null and clearing all data from the line")
    self.sendCommand(CommandNull())
    old_timeout = self.timeout
    self.timeout = timeout
    self.stream.read(None)
    self.stream.reset_input_buffer()
    self.buffer = b''
    self.timeout = old_timeout
    self.current_command = []
    log.info("return")




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

class MicroSpecException(Exception):
  pass

class MicroSpecEmulationException(MicroSpecException):
  pass

class MicroSpecConnectionException(MicroSpecException):
  pass




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.internal.jsonparse import enclassJsonFile
from microspeclib.logger             import CHROMASPEC_LOGGER_DATA as log
import os, sys

CHROMASPEC_COMMAND_ID, CHROMASPEC_COMMAND_NAME = enclassJsonFile("microspec.json", "command")

globals().update([[v.name,v] for k,v in CHROMASPEC_COMMAND_ID.items()])

__all__ = list(CHROMASPEC_COMMAND_NAME.keys())+["getCommandByID","getCommandByName"]

def getCommandByID(cid):
  log.info("cid=%d", cid)
  com = CHROMASPEC_COMMAND_ID.get(cid)
  log.info("return %s", com)
  return com

def getCommandByName(name):
  log.info("name=%s", name)
  com = CHROMASPEC_COMMAND_NAME.get(name)
  log.info("return %s", com)
  return com




import os

__copyright__ = """Copyright 2020 Chromation, Inc"""
__license__   = """All Rights Reserved by Chromation, Inc"""
__doc__       = """
see API documentation: 'python -m pydoc microspeclib.simple'
"""


# NOTE: Sphinx ignores __init__.py files, so for generalized documentation,
#       please use pydoc, or the sphinx-generated documents in doc/build,
#       or the README.md file

# NOTE on CHROMASPEC_ROOTDIR
#
# It is specifically located in the __init__.py of the base microspeclib
# package, so that the ../ (src) ../ (microspec) directory can be found,
# so that, in turn, the cfg and other directories can be referenced
# programmatically and without relative references throughout the
# packages. The test system can find root package directories, but the
# runtime system has no standard for this, and we are avoiding utilizing
# a test system for runtime use.
#
# If microspeclib is in /foo/bar/microspec/src/microspeclib then
# CHROMASPEC_ROOTDIR will be /foo/bar/microspec

CHROMASPEC_ROOTDIR = os.path.realpath(
                       os.path.join(
                         os.path.dirname(__file__), # microspeclib
                         "..",                        # src
                         ".."                         # microspec
                       )
                     )




#!/usr/bin/env python

'''Examples using the low-level version of the API.

USAGE
-----
After "pip install microspec", this script runs from the command
line with the command "microspec-lowlevel-api-example"

$ microspec-lowlevel-api-example
'''

from microspeclib.datatypes import *

def main():
    from microspeclib.expert    import MicroSpecExpertInterface

    xi = MicroSpecExpertInterface(timeout=0.1)
    #xi = MicroSpecExpertInterface(timeout=0.1, device="COM3")
    #xi = MicroSpecExpertInterface(timeout=0.1, device="/dev/cu.usbserial-CHROMATION09310")
    #xi = MicroSpecExpertInterface(timeout=0.1, emulation=True)

    xi.sendCommand(CommandSetBridgeLED(led_num=0, led_setting=LEDOff))
    print(xi.receiveReply())
    xi.sendCommand(CommandSetSensorLED(led_num=0, led_setting=LEDGreen))
    print(xi.receiveReply())
    xi.sendCommand(CommandSetSensorLED(led_num=1, led_setting=LEDRed))
    print(xi.receiveReply())

    xi.sendCommand(CommandGetBridgeLED(led_num=0))
    xi.sendCommand(CommandGetSensorLED(led_num=0))
    xi.sendCommand(CommandGetSensorLED(led_num=1))
    print(xi.receiveReply())
    print(xi.receiveReply())
    reply = xi.receiveReply()
    print(reply.status)
    print(reply.led_setting)

    command = CommandSetSensorConfig(binning=True, gain=Gain1x, row_bitmap=0x1F)
    print(xi.sendAndReceive(command))

    command = CommandSetExposure()
    command.cycles = 100
    print(xi.sendAndReceive(command))

    import time
    for i in range(0,5):
        reply = xi.sendAndReceive(CommandCaptureFrame())
        print(reply.status)
        print(reply.num_pixels)
        pixels = [str(pixel) for pixel in reply.pixels]
        print("first 3 pixels: %s"%(",".join(pixels[:3])))
        time.sleep(0.5)

if __name__ == "__main__":
    main()





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib        import CHROMASPEC_ROOTDIR
from microspeclib.logger import CHROMASPEC_LOGGER_JSON as log
import os, sys

def findConfig(filename):
  log.info("filename=%s", filename)
  developer = os.path.join(CHROMASPEC_ROOTDIR, "cfg", filename)
  installed = os.path.join(sys.prefix, "share", "microspeclib", "cfg", filename)
  local     = os.path.join(".", filename)
  found     = None
  if(  os.access(developer, os.R_OK)): found = developer;
  elif(os.access(installed, os.R_OK)): found = installed;
  elif(os.access(local,     os.R_OK)): found = local;
  else: raise Exception("Cannot find config file %s"%(filename))
  return found





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.expert              import MicroSpecExpertInterface
from microspeclib.logger              import CHROMASPEC_LOGGER as log
from microspeclib.datatypes.command   import CHROMASPEC_COMMAND_NAME
from microspeclib.internal.docstrings import CHROMASPEC_DYNAMIC_DOC
import sys

# The Simple interface doesn't retuire creating objects or doing any sending and waiting
# loops, instead it simply acts like a hardware object that you query for information

# The names of the methods are the names of the Command objects, but with the first letter
# lowercased. You can also look at the names of the commands in the JSON cfg file, or the
# pydoc for this class.

# NOTE: This class is entirely auto-generated, you won't find the API here. Please refer to
#       the Sphinx documentation.

__all__ = [ "MicroSpecSimpleInterface" ]

def _generateFunction(command):
  cname = command.__name__
  name  = cname[7:8].lower()+cname[8:]
  kwargs_param = "".join(["%s=None, "%(v  ) for v in command.variables if v != "command_id"])
  kwargs_data  = "".join(["%s=%s, "  %(v,v) for v in command.variables if v != "command_id"])

  # Generate the parameter list for the function, so that later, documentation introspection finds it properly
  code = """def func(self, %s**kwargs):
    return self.sendAndReceive(command(%s**kwargs))
  """ % (kwargs_param, kwargs_data)
  scope = locals().copy()
  exec(code, scope)
  func = scope["func"]
  func.__qualname__ = "%s.%s"%("MicroSpecSimpleInterface", name)
  func.__doc__      = CHROMASPEC_DYNAMIC_DOC["command"].get(cname, None)

  return name, func

_MicroSpecSimpleInterface = type('_MicroSpecSimpleInterface',
                                 (MicroSpecExpertInterface,),
                                 dict([_generateFunction(command) for command in CHROMASPEC_COMMAND_NAME.values()]))

class MicroSpecSimpleInterface(_MicroSpecSimpleInterface):
  # NOTE: do not docstring this class, the sendAndReceive is overridden but still an underlying, hidden portion
  #       of the API. To see the interface, look at the Sphinx documentation.
  def sendAndReceive(self, command, *args, **kwargs):
    # Simple interface ignores all other processing or in-progress calls, unlike Expert.
    # They are not compatible interfaces, as Expert can be continued after a failed read,
    # but Simple cannot. Do not intermix their usage.
    #
    # It can still get out of sync, but with enough of a wait, it will catch up next time
    log.info("command=%s args=%s kwargs=%s", command, args, kwargs)
    self.stream.reset_input_buffer()
    self.buffer = b''
    self.current_command = []
    reply = super().sendAndReceive(command, *args, **kwargs)
    self.stream.reset_input_buffer()
    self.buffer = b''
    self.current_command = []
    log.info("return %s", reply)
    return reply




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.internal.jsonparse import globalizeJsonFile
import os, sys

CHROMASPEC_GLOBAL = globalizeJsonFile("microspec.json")

globals().update([[k,v] for k,v in CHROMASPEC_GLOBAL.items()])

__all__ = list(CHROMASPEC_GLOBAL.keys())

__doc__ = "\n\n".join([""".. autoattribute:: microspeclib.datatypes.types.%s"""%(g) for g in __all__])




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os
from microspeclib.internal.emulator import *
from microspeclib.datatypes         import *

class MicroSpecTestEmulator(unittest.TestCase):

  def test_chromationEmulatorSettings(self):
    #TODO: when we make more settings, test them here
    e = MicroSpecEmulator()

  def test_chromationEmulatorDefaults(self):
    e = MicroSpecEmulator()
    self.test_chromationEmulatorCompare(e, e)

  def test_chromationEmulatorNull(self):
    e = MicroSpecEmulator()
    assert e.process(CommandNull()) == []

  def test_chromationEmulatorVerify(self):
    e = MicroSpecEmulator()
    assert e.process(CommandVerify()) == [BridgeVerify(status=StatusOK)]

  def test_chromationEmulatorAutoExposure(self):
    e = MicroSpecEmulator()
    assert e.process(CommandAutoExposure()) == \
                    [BridgeAutoExposure(status=StatusOK),
                     SensorAutoExposure(status=StatusOK, success=0, iterations=0)]

  def test_chromationEmulatorCaptureFrame(self):
    e = MicroSpecEmulator()
    #TODO: need to change this when we make the capture frame emulation better
    assert e.process(CommandCaptureFrame()) == \
                    [BridgeCaptureFrame(status=StatusOK),
                     SensorCaptureFrame(status=StatusOK, num_pixels=4, pixels=[111,222,333,444])]

  def test_chromationEmulatorReset(self):
    e1 = MicroSpecEmulator()
    e2 = MicroSpecEmulator()
    self.test_chromationEmulatorSet(e2, bled0=LEDGreen, sled0=LEDRed, sled1=LEDGreen,
                                        binning=True, gain=Gain2_5x, rows=0x15, cycles=2345)
    assert e2.process(CommandReset()) == [BridgeReset(status=StatusOK)]
    self.test_chromationEmulatorCompare(e2, e1)

  def test_chromationEmulatorCompare(self, emulator=None, control=None):
    e = emulator or MicroSpecEmulator()
    c = control  or emulator or MicroSpecEmulator()
    assert e.process(CommandGetBridgeLED(led_num=0)) == \
                    [BridgeGetBridgeLED(status=StatusOK, led_num=0, led_setting=c.bridge_led[0])]
    assert e.process(CommandGetSensorLED(led_num=0)) == \
                    [BridgeGetSensorLED(status=StatusOK),
                     SensorGetSensorLED(status=StatusOK, led_num=0, led_setting=c.sensor_led[0])]
    assert e.process(CommandGetSensorLED(led_num=1)) == \
                    [BridgeGetSensorLED(status=StatusOK),
                     SensorGetSensorLED(status=StatusOK, led_num=1, led_setting=c.sensor_led[1])]
    assert e.process(CommandGetSensorConfig()) == \
                    [BridgeGetSensorConfig(status=StatusOK),
                     SensorGetSensorConfig(status=StatusOK, binning=c.binning, gain=c.gain, row_bitmap=c.rows)]
    assert e.process(CommandGetExposure()) == \
                    [BridgeGetExposure(status=StatusOK),
                     SensorGetExposure(status=StatusOK, cycles=c.cycles)]

  def test_chromationEmulatorSet(self, emulator=None, bled0=LEDOff, sled0=LEDOff, sled1=LEDOff,
                                       binning=BinningDefault, gain=GainDefault, rows=RowsDefault, cycles=0):
    e = emulator or MicroSpecEmulator()
    assert e.process(CommandSetBridgeLED(led_num=0, led_setting=bled0)) == \
                    [BridgeSetBridgeLED(status=StatusOK)]
    assert e.process(CommandSetSensorLED(led_num=0, led_setting=sled0)) == \
                    [BridgeSetSensorLED(status=StatusOK),
                     SensorSetSensorLED(status=StatusOK)]
    assert e.process(CommandSetSensorLED(led_num=1, led_setting=sled1)) == \
                    [BridgeSetSensorLED(status=StatusOK),
                     SensorSetSensorLED(status=StatusOK)]
    assert e.process(CommandSetSensorConfig(binning=binning, gain=gain, row_bitmap=rows)) == \
                    [BridgeSetSensorConfig(status=StatusOK),
                     SensorSetSensorConfig(status=StatusOK)]
    assert e.process(CommandSetExposure(cycles=cycles)) == \
                    [BridgeSetExposure(status=StatusOK),
                     SensorSetExposure(status=StatusOK)]
    assert e.bridge_led[0] == bled0
    assert e.sensor_led[0] == sled0
    assert e.sensor_led[1] == sled1
    assert e.binning       == binning
    assert e.gain          == gain
    assert e.rows          == rows
    assert e.cycles        == cycles

  def test_chromationEmulatorBadData(self):
    e = MicroSpecEmulator()
    assert e.process(CommandGetBridgeLED(led_num=-1)) == \
                    [BridgeGetBridgeLED(status=StatusError, led_num=0, led_setting=LEDOff)]
    assert e.process(CommandGetBridgeLED(led_num=1)) == \
                    [BridgeGetBridgeLED(status=StatusError, led_num=0, led_setting=LEDOff)]
    assert e.process(CommandGetSensorLED(led_num=-1)) == \
                    [BridgeGetSensorLED(status=StatusOK),
                     SensorGetSensorLED(status=StatusError, led_num=0, led_setting=LEDOff)]
    assert e.process(CommandGetSensorLED(led_num=2)) == \
                    [BridgeGetSensorLED(status=StatusOK),
                     SensorGetSensorLED(status=StatusError, led_num=0, led_setting=LEDOff)]
    assert e.process(CommandSetBridgeLED(led_num=-1, led_setting=LEDOff)) == \
                    [BridgeSetBridgeLED(status=StatusError)]
    assert e.process(CommandSetBridgeLED(led_num=1, led_setting=LEDOff)) == \
                    [BridgeSetBridgeLED(status=StatusError)]
    assert e.process(CommandSetBridgeLED(led_num=0, led_setting=99)) == \
                    [BridgeSetBridgeLED(status=StatusError)]
    assert e.process(CommandSetSensorLED(led_num=-1, led_setting=LEDOff)) == \
                    [BridgeSetSensorLED(status=StatusOK),
                     SensorSetSensorLED(status=StatusError)]
    assert e.process(CommandSetSensorLED(led_num=2, led_setting=LEDOff)) == \
                    [BridgeSetSensorLED(status=StatusOK),
                     SensorSetSensorLED(status=StatusError)]
    assert e.process(CommandSetSensorLED(led_num=2, led_setting=99)) == \
                    [BridgeSetSensorLED(status=StatusOK),
                     SensorSetSensorLED(status=StatusError)]
    assert e.process(CommandSetSensorConfig(binning=-1, gain=GainDefault, row_bitmap=RowsDefault)) == \
                    [BridgeSetSensorConfig(status=StatusOK),
                     SensorSetSensorConfig(status=StatusError)]
    assert e.process(CommandSetSensorConfig(binning=True, gain=99, row_bitmap=RowsDefault)) == \
                    [BridgeSetSensorConfig(status=StatusOK),
                     SensorSetSensorConfig(status=StatusError)]
    assert e.process(CommandSetSensorConfig(binning=True, gain=GainDefault, row_bitmap=0x00)) == \
                    [BridgeSetSensorConfig(status=StatusOK),
                     SensorSetSensorConfig(status=StatusError)]
    assert e.process(CommandSetSensorConfig(binning=True, gain=GainDefault, row_bitmap=0x8000)) == \
                    [BridgeSetSensorConfig(status=StatusOK),
                     SensorSetSensorConfig(status=StatusError)]
    assert e.process(CommandSetExposure(cycles=-1)) == \
                    [BridgeSetExposure(status=StatusOK),
                     SensorSetExposure(status=StatusError)]




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest
from timeit import default_timer as timer
from test_expert_interface    import MicroSpecTestExpertInterface
from microspeclib.expert     import MicroSpecExpertInterface
from microspeclib.datatypes  import *
from microspeclib.exceptions import *

@pytest.mark.xfail(raises=MicroSpecConnectionException, strict=False, reason="Hardware not connected")
class MicroSpecTestExpertInterfaceHardware(MicroSpecTestExpertInterface):
  __test__ = True

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.hardware = None
    if not hasattr(cls, "software"):
      cls.software = MicroSpecExpertInterface(timeout=0.1)

  @classmethod
  def tearDownClass(cls):
    super().tearDownClass()
    if hasattr(cls, "software"):
      del cls.software














# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.logger import CHROMASPEC_LOGGER_UTIL as log
from struct import unpack, pack
import re
import itertools

class MicroSpecInteger(int):

  """A subclass of Python int that remembers it's byte width and name so that it can be 
  converted to bytes and str properly"""

  def __new__(self, value, size=1, byteorder="big", signed=False, name=None):
    log.info("value=%d size=%d byteorder=%s signed=%s name=%s", value, size, byteorder, signed, name)
    self = int.__new__(MicroSpecInteger, value)
    self.size      = size
    self.byteorder = byteorder
    self.signed    = signed
    self.name      = name
    log.info("return %s", self)
    return self

  def __str__(self):
    if self.name:
      return self.name
    return super().__str__()

  def __bytes__(self):
    log.info("")
    b = self.to_bytes(self.size, self.byteorder, signed=self.signed)
    log.info("return %s", b)
    return b

def isInt(i):
  log.info("int=%s", i)
  try:
    if int(i) == i:
      log.info("return True")
      return True
  except:
    log.info("return False")
    return False
  return False

def dehex(value):
  log.info("value=%s", value)
  if re.match('^0x[0-9a-fA-F]+$', str(value)):
    h = int(value, 16)
    log.info("return %d", h)
    return h
  elif re.match('^[Tt](?:rue)?$', str(value)):
    log.info("return True")
    return True
  elif re.match('^[Ff](?:alse)?$', str(value)):
    log.info("return False")
    return False
  log.info("return %s", value)
  return value

def _flatten(_list):
  for item in _list:
    if isinstance(item, list) or isinstance(item, tuple):
      yield from flatten(item)
    else:
      yield item

def flatten(_list):
  return list(_flatten(_list))




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.logger import CHROMASPEC_LOGGER_PAYLOAD as log
from .util       import *
from .docstrings import *
from struct      import unpack, pack
import itertools
import re

class MicroSpecPayload(object):
  def _init(self, payload=None, **kwargs):
    log.info("payload=%s kwargs=%s", payload, kwargs)
    self.varsize = {}
    for n in range(0, len(self.variables)):
      log.debug("n=%d", n)
      var   = self.variables[n]
      value = kwargs.get(var, None)
      self.varsize[var] = self.sizes[n]
      setattr(self, var, value)
      log.debug("value[%s]=%s size[%s]=%d", var, value, var, self.sizes[n])
    if payload:
      self.unpack(payload)
    log.info("return")

  def __getitem__(self, attr):
    return getattr(self, attr)

  def __setitem__(self, attr, value):
    setattr(self, attr, value)

  def __setattr__(self, attr, value):
    log.info("attr=%s value=%s", attr, value)
    if attr in self.const:
      pass
    elif attr in self.variables and value is not None:
      self.__dict__.update({attr: MicroSpecInteger(dehex(value), self.varsize[attr])})
    else:
      self.__dict__[attr] = value
    log.info("return")

  def __iter__(self):
    log.info("return")
    return iter(self.variables)

  def __repr__(self):
    log.info("")
    s = "<%s name=%s command_id=%s variables=%s values=%s sizes=%s packformat=%s length=%d packed=%s>" % \
      (self.__class__.__name__, self.name, self.command_id, self.variables, \
        self.values(), self.sizes, self.packformat(), \
        len(self), self.pack())
    log.info("return %s", s)
    return s

  def __str__(self):
    log.info("")
    s = "%s(%s)" % (self.__class__.__name__, \
        ", ".join(["%s=%s"%(k, v) for k, v in self.values().items()]))
    log.info("return %s", s)
    return s

  def csv(self):
    log.info("")
    s = "%s,%s"%(self.__class__.__name__, \
          ",".join([ str(s) for s in flatten(self.values().items()) ])
        )
    log.info("return %s", s)
    return s

  def __bytes__(self):
    log.info("")
    b = self.pack()
    log.info("return %s", b)
    return b

  def __len__(self):
    log.info("")
    l = sum(self.sizes)
    log.info("return %s", l)
    return l

  def values(self):
    log.info("")
    v = {k: getattr(self, k) for k in self.variables}
    log.info("return %s", v)
    return v
    
  def packformat(self):
    log.info("")
    packformat = ">"
    for v in self.variables:
      log.debug("packformat=%s varsize[%s]=%d", packformat, v, self.varsize[v])
      packformat += "B" if self.varsize[v] == 1 else \
                    "H" if self.varsize[v] == 2 else \
                    "L" if self.varsize[v] == 4 else \
                    str(  self.varsize[v]) + "s"
    log.info("return %s", packformat)
    return packformat

  def packvalues(self):
    log.info("")
    pv = [getattr(self, v) for v in self.variables]
    log.info("return %s", pv)
    return pv

  def pack(self):
    log.info("")
    if None in self.values().values():
      log.warning("Marshalling a payload that is missing values, returning ''");
      return b''
    p = pack(self.packformat(), *self.packvalues())
    log.info("return %s", p)
    return p

  def unpack(self, payload):
    log.info("payload=%s", payload)
    # The payload may be bigger than needed, to accommodate
    #   chopping data down packet by packet
    length = len(self)
    values = unpack(self.packformat(), payload[0:length])
    log.debug("unpacked values=%s", values)
    for n in range(0, len(values)):
      log.debug("variables[%d]=%s <- values[%d]=%s", n, self.variables[n], n, values[n])
      setattr(self, self.variables[n], values[n])
    p = payload[length:]
    log.info("return %s", p)
    return p

  def __eq__(self, other):
    if not isinstance(other, MicroSpecPayload):
      return NotImplemented
    if self.name       != other.name:       return False
    if self.command_id != other.command_id: return False
    if self.variables  != other.variables:  return False
    if self.values()   != other.values():   return False
    if self.sizes      != other.sizes:      return False
    if self.varsize    != other.varsize:    return False
    return True

class MicroSpecRepeatPayload(MicroSpecPayload):
  """The difference from the parent class is that the repeat process
  requires packing arrays, and requires partially-gradually unpacking
  the payload, since part of the payload defines how much to continue
  to pack and unpack."""

  def _init(self, *args, **kwargs):
    self.__dict__["repeat"] = self.__class__.repeat.copy()
    super()._init(*args,**kwargs)

  def __setattr__(self, attr, value):
    log.info("attr=%s value=%s", attr, value)
    if attr in self.const:
      pass
    elif attr in self.variables:
      repeat = self.repeat.get(attr, None)
      if repeat:
        self.__dict__.update({attr:
          [] if value is None else
          [MicroSpecInteger(dehex(v), self.varsize[attr]) for v in value]
       })
      else:
        self.__dict__.update({attr:
          value if value is None else
          MicroSpecInteger(dehex(value), self.varsize[attr])
       })
    else:
      self.__dict__.update({attr: value})
    log.info("return")

  def __setitem__(self, attr, value):
    setattr(self, attr, value)

  def packformat(self):
    log.info("")
    packformat = ">"
    for v in self.variables:
      log.debug("packformat=%s v=%s", packformat, v)
      pf = "B" if self.varsize[v] == 1 else \
           "H" if self.varsize[v] == 2 else \
           "L" if self.varsize[v] == 4 else \
           str(  self.varsize[v]) + "s"
      repeat = self.repeat.get(v, None)
      log.debug("repeat=%s", repeat)
      if repeat:
        repeat_num = self[repeat]
        log.debug("repeat_num=%s", repeat_num)
        pf = pf * int(repeat_num if repeat_num is not None else 1)
      packformat += pf
    log.info("return %s", packformat)
    return packformat

  def packvalues(self):
    log.info("")
    pv = list(itertools.chain.from_iterable(
             [  getattr(self, v)[0:int(self[self.repeat.get(v)])] 
               if self.repeat.get(v,None) else
                 [getattr(self, v)]
               for v in self.variables]
          ))
    log.info("return %s", pv)
    return pv

  def unpack(self, payload): 
    log.info("payload=%s",payload)
    # The payload may be bigger than needed, to accommodate
    # chopping data down packet by packet
    packformat = super().packformat()
    # The superclass packformat doesn't repeat, so it's used to iterate through
    # the elements one at a time, whereas the pack function can pack everything
    # at once
    endian     = packformat[0]
    packrest   = packformat[1:]
    payrest    = payload[:]
    log.debug("packformat=%s payrest=%s",packformat,payrest)
    used   = 0
    for n in range(0, len(self.variables)):
      repeat = self.repeat.get(self.variables[n], None)
      log.debug("packformat=%s payrest=%s n=%d repeat=%s variables[%d]=%s", 
                packformat, payrest, n, repeat, n, self.variables[n])
      if repeat:
        size     = self.sizes[n]
        number   = int(self[repeat])
        packmany = packrest[0] * number
        log.debug("if-repeat size=%d number=%d packmany=%s", size, number, packmany)
        value    = unpack(endian+packmany, payrest[0:size*number])
        log.debug("value=%s", value)
        used    += size*number
        packrest = packrest[1:]
        payrest  = payrest[size*number:]
        log.debug("used=%d packrest=%s payrest=%s", used, packrest, payrest)
        setattr(self, self.variables[n], value)
      else:
        size     = self.sizes[n]
        log.debug("if-no-repeat size=%d", size)
        value    = unpack(endian+packrest[0], payrest[0:size])
        log.debug("value=%s", value)
        used    += size
        packrest = packrest[1:]
        payrest  = payrest[size:]
        log.debug("used=%d packrest=%s payrest=%s", used, packrest, payrest)
        setattr(self, self.variables[n], value[0])
    p = payload[used:]
    log.info("return %s", p)
    return p

  def __eq__(self, other):
    if not isinstance(other, MicroSpecRepeatPayload):
      log.info("return NotImplemented")
      return NotImplemented
    if not super().__eq__(other):   
      log.info("return False")
      return False
    if self.repeat != other.repeat: 
      log.info("return False")
      return False
    log.info("return True")
    return True

def MicroSpecPayloadClassFactory(protocol, command_id, name, variables, sizes, repeat=None):
  log.info("protocol=%s command_id=%d name=%s variables=%s sizes=%s repeat=%s", 
           protocol, command_id, name, variables, sizes, repeat)

  # First, form Voltron - er, the base class
  #
  # The classes don't differ based on protocol (i.e. command vs bridge vs sensor)
  #
  # However, they do need to pull out the variables and add them to the list of
  # attributes, hence the **dict(...for v in variables...)
  if repeat:
    klass = type(str(name), (MicroSpecRepeatPayload,), dict({
      'command_id'   : int(command_id),
      'name'         : name,
      'variables'    : variables,
      'sizes'        : sizes,
      'repeat'       : repeat,
      'const'        : ['command_id', 'name', 'variables', 'sizes', 'repeat', 'const'],
    }, **dict([[v, None] for v in variables if v != "command_id"])))
  else:
    klass = type(str(name), (MicroSpecPayload,), dict({
      'command_id'   : int(command_id),
      'name'         : name,
      'variables'    : variables,
      'sizes'        : sizes,
      'const'        : ['command_id', 'name', 'variables', 'sizes', 'const'],
    }, **dict([[v, None] for v in variables if v != "command_id"])))

  # Then create a custom __init__ function with the proper optional parameters, so
  # that documentation can pull out data via introspection later
  kwarg_param = "".join(["%s=None, "%(v) for v in variables if v != "command_id"])
  kwarg_data  = "".join(["%s=%s, "%(v,v) for v in variables if v != "command_id"])
  scope = locals().copy()
  exec("""def __init__(self, *args, %s**kwargs):
    self._init(*args, %s**kwargs)
  """%(kwarg_param, kwarg_data), scope)
  klass.__init__ = scope["__init__"]
  klass.__init__.__qualname__ = "%s.__init__"%(name)

  # And finally insert the docstrings from the internal library, taking care to not
  # assume the docs exist
  klass.__doc__ = CHROMASPEC_DYNAMIC_DOC.get(protocol, {}).get(str(name), "")

  log.info("return %s", klass)
  return klass




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import logging

CHROMASPEC_LOG_FORMAT="%(asctime)-15s:%(filename)s:%(funcName)s:%(lineno)d: %(message)s"
logging.basicConfig(format=CHROMASPEC_LOG_FORMAT)

CHROMASPEC_LOGGER         =logging.getLogger("microspeclib")
CHROMASPEC_LOGGER_TEST    =logging.getLogger("microspeclib.test")
CHROMASPEC_LOGGER_INTERNAL=logging.getLogger("microspeclib.internal")
CHROMASPEC_LOGGER_STREAM  =logging.getLogger("microspeclib.internal.stream")
CHROMASPEC_LOGGER_UTIL    =logging.getLogger("microspeclib.internal.util")
CHROMASPEC_LOGGER_PAYLOAD =logging.getLogger("microspeclib.internal.payload")
CHROMASPEC_LOGGER_JSON    =logging.getLogger("microspeclib.internal.json")
CHROMASPEC_LOGGER_DATA    =logging.getLogger("microspeclib.internal.data")

CHROMASPEC_LOGGER         .setLevel(logging.ERROR)
CHROMASPEC_LOGGER_TEST    .setLevel(logging.ERROR)
CHROMASPEC_LOGGER_INTERNAL.setLevel(logging.ERROR)

def verbose(includeInternals=False):
  """Turn on logging.

  Parameters
  ----------
  includeInternals: True or False
    Also activate logging for internal modules

  """
  CHROMASPEC_LOGGER           .setLevel(logging.WARNING)
  if includeInternals:
    CHROMASPEC_LOGGER_TEST    .setLevel(logging.WARNING)
    CHROMASPEC_LOGGER_INTERNAL.setLevel(logging.WARNING)
  
def quiet(includeInternals=False):
  """Turn off logging.

  Parameters
  ----------
  includeInternals: True or False
    Also deactivate logging for internal modules

  """
  CHROMASPEC_LOGGER           .setLevel(logging.ERROR)
  if includeInternals:
    CHROMASPEC_LOGGER_TEST    .setLevel(logging.ERROR)
    CHROMASPEC_LOGGER_INTERNAL.setLevel(logging.ERROR)
  
def debug(includeInternals=False):
  """Turn on verbose debugging trace.

  Parameters
  ----------
  includeInternals: True or False
    Also activate verbose debugging for internal modules

  """
  CHROMASPEC_LOGGER           .setLevel(logging.DEBUG)
  if includeInternals:
    CHROMASPEC_LOGGER_TEST    .setLevel(logging.DEBUG)
    CHROMASPEC_LOGGER_INTERNAL.setLevel(logging.DEBUG)
  
# Level uses:
# DEBUG:   Every substantial loop and branch
# INFO:    Every subroutine call and return
# WARNING: Minor errors that can usually be ignored
# ERROR:   Data format or API errors
# CRITIAL: Fatal errors like unopenable files or command line argument errors

# If you set the main logging level, everything will print, but if you set it
# just for one component, then just those and below will be produced

# One level per file isn't sufficient, as the reply and command data modules
# are spread out in multiple files





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest
from io import BytesIO
from microspeclib.internal.stream   import MicroSpecBytesIOStream
from microspeclib.datatypes         import *
from microspeclib.datatypes.command import CHROMASPEC_COMMAND_ID
from microspeclib.logger            import CHROMASPEC_LOGGER_TEST as log

class MicroSpecTestBytesIOStream(unittest.TestCase):

  def generate_streams(self):
    b = BytesIO()
    s = MicroSpecBytesIOStream(stream=b)
    return b, s

  def test_defaultStream(self):
    b, s = self.generate_streams()
    s = MicroSpecBytesIOStream()
    assert b is not s.stream

  def test_parameterStream(self):
    b, s = self.generate_streams()
    assert b is s.stream

  def test_underlyingStreamRead1(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02'
    self.write_underlying(b, s, d)
    r = s.read(1)
    assert d[0] == r[0]
    assert len(r) == 1

  def test_underlyingStreamRead1Fail(self):
    b, s = self.generate_streams()
    d = b''
    self.write_underlying(b, s, d)
    r = s.read(1)
    assert len(r) == 0

  def test_underlyingStreamReadAll(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02'
    self.write_underlying(b, s, d)
    r = s.read(3)
    assert     r  ==     d
    assert len(r) == len(d)

  def test_underlyingStreamReadManyFail(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02'
    self.write_underlying(b, s, d)
    r = s.read(10)
    assert     r  ==     d
    assert len(r) == len(d)

  def test_underlyingStreamWrite(self):
    b, s = self.generate_streams()
    b = BytesIO()
    s = MicroSpecBytesIOStream(stream=b)
    d = b'\x00\x01\x02'
    self.write_direct(b, s, d)
    self.seek( s, 0 )
    w = s.read()
    assert     w  ==     d
    assert len(w) == len(d)

  def test_write1Read1(self):
    b, s = self.generate_streams()
    d = b'\x00'
    self.write_direct(b, s, d)
    r = s.read(1)
    assert     r  ==     d
    assert len(r) == len(d)
    # No consume() means no moving forwards
    r = s.read(1)
    assert     r  ==     d
    assert len(r) == len(d)

  def test_write1Read1Consume1(self):
    b, s = self.generate_streams()
    d = b'\x00'
    self.write_direct(b, s, d)
    r = s.read(1)
    assert     r  ==     d
    assert len(r) == len(d)
    s.consume(1)
    r = s.read(1)
    assert     r  ==     b''
    assert len(r) == len(b'')

  def test_writeManyRead1(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02'
    self.write_direct(b, s, d)
    r = s.read(1)
    assert     r  ==     d[0:1]
    assert len(r) == len(d[0:1])
    s.consume(1)
    r = s.read(1)
    assert     r  ==     d[1:2]
    assert len(r) == len(d[1:2])
    s.consume(1)
    r = s.read(1)
    assert     r  ==     d[2:3]
    assert len(r) == len(d[2:3])

  def test_writeManyReadInterleaved(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02\x03\x04\x05'
    self.write_direct(b, s, d[0:3])
    r = s.read(1)
    assert     r  ==     d[0:1]
    assert len(r) == len(d[0:1])
    self.write_direct(b, s, d[3:4])
    s.consume(1)
    r = s.read(1)
    assert     r  ==     d[1:2]
    assert len(r) == len(d[1:2])
    s.consume(1)
    self.write_direct(b, s, d[4:5])
    r = s.read(1)
    assert     r  ==     d[2:3]
    assert len(r) == len(d[2:3])
    s.consume(1)
    r = s.read(1)
    self.write_direct(b, s, d[5:6])
    assert     r  ==     d[3:4]
    assert len(r) == len(d[3:4])
    s.consume(1)
    r = s.read(1)
    assert     r  ==     d[4:5]
    assert len(r) == len(d[4:5])
    s.consume(1)
    r = s.read(1)
    assert     r  ==     d[5:6]
    assert len(r) == len(d[5:6])

  def test_writeManyReadPartialConsume(self):
    b, s = self.generate_streams()
    d = b'\x00\x01\x02\x03\x04\x05'
    self.write_direct(b, s, d)
    r = s.read(4)
    assert     r  ==     d[0:4]
    assert len(r) == len(d[0:4])
    s.consume(2)
    r = s.read(4)
    assert     r  ==     d[2:6]
    assert len(r) == len(d[2:6])
    s.consume(2)
    r = s.read(4)
    assert     r  ==     d[4:6]
    assert len(r) == len(d[4:6])
    s.consume(2)
    r = s.read(4)
    assert     r  ==     d[6:6]
    assert len(r) == len(d[6:6])

  def test_writeReadCommand(self):
    b, s = self.generate_streams()
    w = []
    for cid in CHROMASPEC_COMMAND_ID.keys():
      if cid < 0: continue # Unimplemented test values in JSON
      cklass = getCommandByID(cid)
      c      = cklass()
      for v in c:
        c[v] = 99 # dummy data
      c.command_id = cklass.command_id # varibles include this, need to undo it
      self.sendCommand(b, s, c)
      w.append(c)
    self.small_wait()
    for cid in CHROMASPEC_COMMAND_ID.keys():
      if cid < 0: continue # Unimplemented test values in JSON
      c1 = s.receiveCommand()
      c2 = w.pop(0)
      assert c1 == c2

  def test_partialReadCommand(self):
    b, s = self.generate_streams()
    c = CommandGetBridgeLED(led_num=0)
    self.assert_getvalue(s.stream, b'')

    cb = bytes(c)
    cb1 = cb[0:len(cb)-1]
    cb2 = cb[len(cb)-1:]
    self.write_direct(b, s, cb1)
    self.assert_getvalue(s.stream, cb1)
    self.assert_readpos( s,        0  )
    assert s.buffer            == b''

    self.small_wait()
    r = s.receiveCommand()
    self.assert_getvalue(s.stream, cb1)
    self.assert_readpos( s,        1  )
    assert s.buffer            == cb1
    assert r is None

    self.write_direct(b, s, cb2)
    self.assert_getvalue(s.stream, cb1+cb2)
    self.assert_readpos( s,        1      )
    assert s.buffer            == cb1

    self.small_wait()
    r = s.receiveCommand()
    self.assert_getvalue(s.stream, cb1+cb2)
    self.assert_readpos( s,        2      )
    assert s.buffer            == b''
    assert r == CommandGetBridgeLED(led_num=0)

  def test_partialReadBridgeReply(self):
    b, s = self.generate_streams()
    r = BridgeGetBridgeLED(status=0, led_num=0, led_setting=1)
    self.assert_getvalue(s.stream, b'')

    rb = bytes(r)
    rb1 = rb[0:len(rb)-1]
    rb2 = rb[len(rb)-1:]
    self.write_direct(b, s, rb1)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        0  )
    assert s.buffer            == b''

    self.small_wait()
    x = s.receiveReply(r.command_id)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        1  )
    assert s.buffer            == rb1
    assert x is None

    self.write_direct(b, s, rb2)
    self.assert_getvalue(s.stream, rb1+rb2)
    self.assert_readpos( s,        1      )
    assert s.buffer            == rb1

    self.small_wait()
    x = s.receiveReply(r.command_id)
    self.assert_getvalue(s.stream, rb1+rb2)
    self.assert_readpos( s,        2      )
    assert s.buffer+s.read(0)  == b''
    assert x == BridgeGetBridgeLED(status=0, led_num=0, led_setting=1)

  def test_partialReadBridgeAndSensorReply(self):
    b, s = self.generate_streams()
    r1 = BridgeGetSensorLED(status=0)
    r2 = SensorGetSensorLED(status=1, led_setting=2)
    self.assert_getvalue(s.stream, b'')

    r1b = bytes(r1)
    r2b = bytes(r2)
    rb = r1b + r2b
    rb1 = rb[0:len(rb)-1]
    rb2 = rb[len(rb)-1:]
    self.write_direct(b, s, rb1)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        0  )
    assert s.buffer            == b''

    self.small_wait()
    x = s.receiveReply(r1.command_id)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        2  )
    assert s.buffer            == rb1
    assert x is None

    self.write_direct(b, s, rb2)
    self.assert_getvalue(s.stream, rb1+rb2)
    self.assert_readpos( s,        2      )
    assert s.buffer            == rb1

    self.small_wait()
    x = s.receiveReply(r1.command_id)
    self.assert_getvalue(s.stream, rb1+rb2)
    self.assert_readpos( s,        3      )
    assert s.buffer            == b''
    assert x == SensorGetSensorLED(status=1, led_setting=2)

  def test_partialReadBridgeNonzerostatusAndSensorReply(self):
    b, s = self.generate_streams()
    r1 = BridgeGetSensorLED(status=1)
    r2 = SensorGetSensorLED(status=0, led_setting=2)
    self.assert_getvalue(s.stream, b'')

    r1b = bytes(r1)
    r2b = bytes(r2)
    rb = r1b + r2b
    rb1 = rb[0:len(rb)-1]
    rb2 = rb[len(rb)-1:]
    self.write_direct(b, s, rb1)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        0  )
    assert s.buffer            == b''

    self.small_wait()
    x = s.receiveReply(r1.command_id)
    self.assert_getvalue(s.stream, rb1)
    self.assert_readpos( s,        2  )
    assert s.buffer            == rb1[1:]
    assert x == BridgeGetSensorLED(status=1)

  def test_writeReadReply(self):
    b, s = self.generate_streams()
    w = []
    for cid in CHROMASPEC_COMMAND_ID.keys():
      if cid < 0: continue # Unimplemented test values in JSON
      r1klass = getBridgeReplyByID(cid)
      r1      = r1klass()
      for v in r1:
        r1[v] = 99 # dummy data
      r1.command_id = r1klass.command_id              # varibles include this, need to undo it
      if r1.__dict__.get("status", None) is not None: # the null replies don't have this
        r1.status   = 0                               # set status 0 otherwise we trigger a different check
      log.info("sending bridge %s"%(r1))
      self.sendReply(b, s, r1)
      w.append(r1)
      r2klass = getSensorReplyByID(cid)
      if r2klass:
        r2      = r2klass()
        for v in r2:
          if r2.__dict__.get("repeat", None) is not None:
            if r2.repeat.get(v):
              r2[v] = [99]*99
              r2[r2.repeat[v]] = 99
            else:
              r2[v] = 99 # dummy data
          else:
            r2[v] = 99 # dummy data
        r2.command_id = r2klass.command_id              # varibles include this, need to undo it
        if r2.__dict__.get("status", None) is not None: # the null replies don't have this
          r2.status   = 0                               # set status 0 otherwise we trigger a different check
        log.info("sending sensor %s"%(r1))
        self.sendReply(b, s, r2)
        w.append(r2)
    self.small_wait()
    for cid in CHROMASPEC_COMMAND_ID.keys():
      if cid < 0: continue # Unimplemented test values in JSON
      r1 = s.receiveReply(cid)
      log.info("received %s"%(r1))
      r2 = w.pop(0)
      log.info("popped %s"%(r2))
      if getSensorReplyByID(cid):
        # If there's a sensor reply, we'll only get that half back with this call,
        #   never the bridge reply portion
        r2 = w.pop(0)
        log.info("popped additional %s"%(r2))
      assert r1 == r2

  # Note: all of these are here so we can override them in the emulator test and change where we're
  # pushing and pulling the data from/to:

  def seek(self, stream, pos):
    return stream.stream.seek(pos)

  def write_underlying(self, b, s, *args, **kwargs):
    return b.write(*args, **kwargs)

  def write_direct(self, b, s, *args, **kwargs):
    return s.write(*args, **kwargs)

  def sendCommand(self, b, s, *args, **kwargs):
    return s.sendCommand(*args, **kwargs)

  def sendReply(self, b, s, *args, **kwargs):
    return s.sendCommand(*args, **kwargs)

  def assert_getvalue(self, stream, value):
    assert stream.getvalue() == value

  def assert_readpos(self, stream, pos):
    assert stream.readpos == pos

  def small_wait(self):
    pass





















# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os
from microspeclib.internal.util import MicroSpecInteger, isInt, dehex

class MicroSpecTestUtil(unittest.TestCase):

  def test_chromationInteger(self):
    ints = [
      [   0, 1,    "big", False,     b'\x00'],
      [   1, 1,    "big", False,     b'\x01'],
      [   1, 2,    "big", False, b'\x00\x01'],
      [   1, 2, "little", False, b'\x01\x00'],
      [-128, 1,    "big",  True,     b'\x80'],
      [ 127, 1,    "big",  True,     b'\x7f'],
      [  -1, 1,    "big",  True,     b'\xff'],
      [ 255, 1,    "big", False,     b'\xff'],
    ]
    for i in ints:
      ci = MicroSpecInteger(i[0], i[1], i[2], i[3])
      assert ci == i[0]
      b = bytes(ci)
      assert b == i[4]
    
  def test_isInt(self):
    assert isInt(1)     == True
    assert isInt("1")   == False
    assert isInt(None)  == False
    #assert isInt(False) == False # These actually turn out to be True
    #assert isInt(True)  == False # These actually turn out to be True

  def test_dehex(self):
    assert dehex("T")     == True
    assert dehex("True")  == True
    assert dehex("F")     == False
    assert dehex("False") == False
    assert dehex("t")     == True
    assert dehex("true")  == True
    assert dehex("f")     == False
    assert dehex("false") == False
    assert dehex("foo")   == "foo"
    assert dehex("FF")    == "FF"
    assert dehex("1")     == "1"
    assert dehex("0xFF")  == 255
    assert dehex("0xff")  == 255




#!/usr/bin/env python

# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

"""
Example Usage
=============

NOTE
----
  You probably never need to run this, unless you really need to turn on debugging trace on the
  emulator in order to see what you're sending to it. Usually, specifying emulator=True is enough
  to use the emulator, in the simple and expert API.

  Also, this will only work on Mac OSX and Linux systems, not on Windows.

Longwinded example
------------------
dir=`mktemp -d`

socat PTY,raw,echo=0,link=$dir/microspec.software PTY,raw,echo=0,link=$dir/microspec.hardware &

microspec_emulator.py -f $dir/microspec.hardware

# Then connect interface to $dir/microspec.software file

# And stop the socat background command, and clean up the $dir and it's contents

Short example
-------------
microspec_emulator.py -s -p

# Then connect interface to filename that the script prints to stdout

"""

def main():
  import sys
  import argparse
  parser = argparse.ArgumentParser(
    description="Create software instance of Chromation hardware for use with testing access software")
  parser.add_argument("-v", "--verbose", help="Print verbose trace",                action="count", default=0)
  parser.add_argument("-d", "--debug",   help="Internal debugging",                 action="count", default=0)
  parser.add_argument("-p", "--print",   help="Print socket to connect to",         action="count", default=0)
  parser.add_argument("-t", "--timeout", help="Timeout in (partial float) seconds", nargs=1,        default=[0],
                                                                                    action="store", type=float)
  parser.add_argument("-s", "--spawn",   help="Spawn socat instance?",              action="count", default=False)
  parser.add_argument("-f", "--file",    help="File to use as a pipe - default create anew and print location", 
                                                                                                    default=None)
  args = parser.parse_args()
  
  from microspeclib.internal.emulator import MicroSpecEmulator
  from microspeclib.internal.stream   import MicroSpecEmulatedStream
  from microspeclib.logger            import debug
  import logging
  
  log = logging.getLogger("microspec_emulator")
  if args.verbose or args.debug: 
    debug( args.debug>0 )
    log.setLevel( logging.DEBUG )
  
  serial   = MicroSpecEmulatedStream(hardware=args.file, timeout=args.timeout[0], socat=args.spawn, fork=False)
  emulator = MicroSpecEmulator()
  
  if args.file is None or args.print:
    print("%s"%(serial.software))
    # Note: necessary so that the emulator object that spawns this program and waits for some input
    # on stdout definitely gets data, otherwise it gets stuck in the buffer and never triggers select:
    sys.stdout.flush() 
  
  while True:
    log.info("Waiting for command...")
    command = serial.receiveCommand()
    if not command:
      log.info("No command found")
      continue
    log.info("Received command %s"%(command))
    reply = emulator.process(command)
    if reply:
      log.info("Sending reply %s"%(reply))
      for packet in reply:
        serial.sendReply(packet)
    else:
      log.info("No reply to send")
  
if __name__ == "__main__":
  main()




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.datatypes import *

class MicroSpecEmulator(object):
  
  def __init__(self):
    #TODO: more initialization state
    self.bridge_led = {0: LEDOff}
    self.sensor_led = {0: LEDOff, 1: LEDOff}
    self.binning    = BinningDefault
    self.gain       = GainDefault
    self.rows       = RowsDefault
    self.cycles     = 0 #TODO: default cycles value
    self.max_tries        = 0
    self.start_pixel      = 0
    self.stop_pixel       = 0
    self.target           = 0
    self.target_tolerance = 0
    self.max_exposure     = 0
    self.success          = 0
    self.iterations       = 0

  def process(self, command):
    #import pdb; pdb.set_trace()
    if command.command_id == CommandGetBridgeLED.command_id:
      try:
        num = command.led_num
        assert 0 <= num <= 0
        return [BridgeGetBridgeLED(led_num=num, led_setting=self.bridge_led[num], status=StatusOK)]
      except:
        return [BridgeGetBridgeLED(led_num=num, led_setting=LEDOff,               status=StatusError)]
    elif command.command_id == CommandSetBridgeLED.command_id:
      try:
        num = command.led_num
        led = command.led_setting
        assert 0 <= num <= 0
        assert led in [LEDOff, LEDGreen, LEDRed]
        self.bridge_led[num] = led
        return [BridgeSetBridgeLED(status=StatusOK)]
      except:
        return [BridgeSetBridgeLED(status=StatusError)]
    elif command.command_id == CommandGetSensorLED.command_id:
      try:
        num = command.led_num
        assert 0 <= num <= 1
        return [BridgeGetSensorLED(                                               status=StatusOK),
                SensorGetSensorLED(led_num=num, led_setting=self.sensor_led[num], status=StatusOK)]
      except:
        return [BridgeGetSensorLED(                                               status=StatusOK),
                SensorGetSensorLED(led_num=num, led_setting=LEDOff,               status=StatusError)]
    elif command.command_id == CommandSetSensorLED.command_id:
      try:
        num = command.led_num
        led = command.led_setting
        assert 0 <= num <= 1
        assert led in [LEDOff, LEDGreen, LEDRed]
        self.sensor_led[num] = led
        return [BridgeSetSensorLED(status=StatusOK),
                SensorSetSensorLED(status=StatusOK)]
      except:
        return [BridgeSetSensorLED(status=StatusOK),
                SensorSetSensorLED(status=StatusError)]
    elif command.command_id == CommandReset.command_id:
      self.bridge_led = {0: LEDOff}
      self.sensor_led = {0: LEDOff, 1: LEDOff}
      self.binning    = BinningDefault
      self.gain       = GainDefault
      self.rows       = RowsDefault
      self.cycles     = 0 #TODO: default cycles value
      return [BridgeReset(status=StatusOK)]
    elif command.command_id == CommandVerify.command_id:
      return [BridgeVerify(status=StatusOK)]
    elif command.command_id == CommandNull.command_id:
      return []
    elif command.command_id == CommandGetSensorConfig.command_id:
      return [BridgeGetSensorConfig(status=StatusOK),
              SensorGetSensorConfig(status=StatusOK, binning=self.binning, gain=self.gain, row_bitmap=self.rows)]
    elif command.command_id == CommandSetSensorConfig.command_id:
      try:
        assert False <= command.binning <= True
        assert command.gain in [Gain1x, Gain2_5x, Gain4x, Gain5x]
        assert command.row_bitmap != 0
        assert command.row_bitmap&0x1F != 0
        self.gain = command.gain
        self.binning = command.binning
        self.rows = command.row_bitmap
        return [BridgeSetSensorConfig(status=StatusOK),
                SensorSetSensorConfig(status=StatusOK)]
      except:
        return [BridgeSetSensorConfig(status=StatusOK),
                SensorSetSensorConfig(status=StatusError)]
    elif command.command_id == CommandAutoExposure.command_id:
      return [BridgeAutoExposure(status=StatusOK),
              SensorAutoExposure(status=StatusOK, success=0, iterations=0)]
    elif command.command_id == CommandGetAutoExposeConfig.command_id:
      return [BridgeGetAutoExposeConfig(status=StatusOK),
              SensorGetAutoExposeConfig(status=StatusOK, max_tries=self.max_tries, start_pixel=self.start_pixel,
                                        stop_pixel=self.stop_pixel, target=self.target, 
                                        target_tolerance=self.target_tolerance, max_exposure=self.max_exposure)]
    elif command.command_id == CommandSetAutoExposeConfig.command_id:
      self.max_tries        = command.max_tries
      self.start_pixel      = command.start_pixel
      self.stop_pixel       = command.stop_pixel
      self.target           = command.target
      self.target_tolerance = command.target_tolerance
      self.max_exposure     = command.max_exposure
      return [BridgeSetAutoExposeConfig(status=StatusOK),
              SensorSetAutoExposeConfig(status=StatusOK)]
    elif command.command_id == CommandGetExposure.command_id:
      return [BridgeGetExposure(status=StatusOK),
              SensorGetExposure(status=StatusOK, cycles=self.cycles)]
    elif command.command_id == CommandSetExposure.command_id:
      try:
        assert 0x00 <= command.cycles <= 0xFFFF
        self.cycles = command.cycles
        return [BridgeSetExposure(status=StatusOK),
                SensorSetExposure(status=StatusOK)]
      except:
        return [BridgeSetExposure(status=StatusOK),
                SensorSetExposure(status=StatusError)]
    elif command.command_id == CommandCaptureFrame.command_id:
      #TODO: big todo - play back recorded or make up data etc
      return [BridgeCaptureFrame(status=StatusOK),
              SensorCaptureFrame(status=StatusOK, num_pixels=4, pixels=[111,222,333,444])]
    return []




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.internal.jsonparse import enclassJsonFile
from microspeclib.logger             import CHROMASPEC_LOGGER_DATA as log
import os, sys

CHROMASPEC_SENSOR_ID, CHROMASPEC_SENSOR_NAME = enclassJsonFile("microspec.json", "sensor")

globals().update([v.name,v] for k,v in CHROMASPEC_SENSOR_ID.items())

__all__ = list(CHROMASPEC_SENSOR_NAME.keys())+["getSensorReplyByID","getSensorReplyByName"]

def getSensorReplyByID(cid):
  log.info("cid=%d", cid)
  com = CHROMASPEC_SENSOR_ID.get(cid)
  log.info("return %s", com)
  return com

def getSensorReplyByName(name):
  log.info("name=%s", name)
  com = CHROMASPEC_SENSOR_NAME.get(name)
  log.info("return %s", com)
  return com





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os
from microspeclib.internal.jsonparse import globalizeJsonFile, enclassJsonFile

jsontest = os.path.realpath(os.path.join(os.path.dirname(__file__), "test.json"))

class MicroSpecTestJson(unittest.TestCase):

  def test_globals(self):
    g = globalizeJsonFile(jsontest)
    assert g["hex"]    == 255
    assert g["int"]    == 256
    assert g["string"] == "foo"
    assert g["bool"]   == False

  def test_command(self):
    cid, cname = enclassJsonFile(jsontest, protocol="command")
    assert cid  [0         ]().name         == "Commanda"
    assert cid  [1         ]().variables[0] == "command_id"
    assert cid  [2         ]().sizes[2]     == 4
    assert cid  [2         ]().repeat["b"]  == "a"
    assert cname["Commanda"]().name         == "Commanda"
    assert cname["Commandb"]().variables[0] == "command_id"
    assert cname["Commandc"]().sizes[2]     == 4
    assert cname["Commandc"]().repeat["b"]  == "a"

  def test_bridge(self):
    cid, cname = enclassJsonFile(jsontest, protocol="bridge")
    assert cid  [0        ]().name         == "Bridgea"
    assert cid  [1        ]().variables[0] == "command_id"
    assert cid  [2        ]().sizes[2]     == 4
    assert cid  [2        ]().repeat["b"]  == "a"
    assert cname["Bridgea"]().name         == "Bridgea"
    assert cname["Bridgeb"]().variables[0] == "command_id"
    assert cname["Bridgec"]().sizes[2]     == 4
    assert cname["Bridgec"]().repeat["b"]  == "a"

  def test_sensor(self):
    cid, cname = enclassJsonFile(jsontest, protocol="sensor")
    assert cid  [0        ]().name         == "Sensora"
    assert cid  [1        ]().variables[0] == "command_id"
    assert cid  [2        ]().sizes[2]     == 4
    assert cid  [2        ]().repeat["b"]  == "a"
    assert cname["Sensora"]().name         == "Sensora"
    assert cname["Sensorb"]().variables[0] == "command_id"
    assert cname["Sensorc"]().sizes[2]     == 4
    assert cname["Sensorc"]().repeat["b"]  == "a"




"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

import glob

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'PyPI.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    include_package_data=True,

    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='microspec',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.1a6',  # Required (reset to this before PyPI publish)
    # version='0.1.1a13',  # Temporary increment for TestPyPI publish)

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Python API for the Chromation spectrometer dev-kit',

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/markdown',

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    # url='https://www.chromation.com/',
    url='https://github.com/microspectrometer/microspec',


    # This should be your name or the name of the organization which owns the
    # project.
    author='Chromation',
    author_email="mike@chromationspec.com",

    # This should be a valid email address corresponding to the author listed
    # above.
    #author_email='pypa-dev@googlegroups.com',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        #'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        # Additional classifiers
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Terminals :: Serial",
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='spectrometer API USB serial hardware embedded-systems',

    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    package_dir={'': 'src'},

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(where='src'),  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.7', # need 3.7 for namedtuple kwarg "defaults"

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['pyserial'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={
        # For developers:
        # emulate hardware (Linux/Mac only, requires socat)
        # run unit tests after modifying API
        # rebuild documentation after modifying API
        'dev': [
            'sphinx', # docs
            'sphinx-rtd-theme', # docs
            'recommonmark', # docs
            'm2r', # docs
            'sphinxcontrib-argdoc', # docs
            'pytest', # docs and unit test
            'pytest-testdox', # pretty-print unit tests
            'tabulate', # docs and unit test
            'psutil', # docs and emulator (requires socat)
            ],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    #package_data={
    #    'sample': ['package_data.dat'],
    #},
    # DO NOT USE package_data AND MANIFEST.in AT THE SAME TIME

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # For example, if installing in virtual environment `foo`,
    # The microspec package is installed in
    # foo\Lib\site-packages\microspeclib
    # But these `data_files` are installed outside the package:
    # - the tests are in foo/share/microspeclib/tests
    # - the docs are in foo/share/microspeclib/docs
    # - the JSON file is in foo/share/microspeclib/cfg
    data_files=[
      ('share/microspeclib/cfg',                     ['cfg/microspec.json']),
      ('share/microspeclib/tests',                   glob.glob("tests/*.*")),
      ('share/microspeclib/doc/build/html',          glob.glob("doc/build/html/*.*")),
      ('share/microspeclib/doc/build/html/_static',  glob.glob("doc/build/html/_static/*.*")),
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    #entry_points={  # Optional
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    #},
    # example: run from PowerShell: `> microspec_simple_example.py`
    # On Windows, creates foo\Scripts\simple_example.py
    scripts=['bin/microspec_cmdline.py',
        'bin/microspec_emulator.py',
        'bin/microspec_expert_example.py',
        'bin/microspec_simple_example.py'
    ],
    # entry_points style requires `bin` is a module and `main` is
    # a function.
    # entry_points={
    #     "console_scripts": [
    #         "simple_example=bin.microspec_simple_example:main",
    #     ],
    # but `bin` is not a module and `main` is not a function.
    # Copy script to new directory `microspeclib/examples`.
    # Rename as 'microspec_api.py'.
    # The script name cannot contain "-", use "_" instead.
    # `import *` is not allowed inside a function call.
    # The entry point key must be "console_scripts". This:
    # - installs the .exe in the {venv}\Scripts folder on Windows
    # - installs the script in the {venv}/bin folder on Linux
    # Run from PowerShell: `> microspec-cmdline`
    # Run from bash: `$ microspec-cmdline`
    # On Windows, install creates {venv}\Scripts\microspec-api-example.exe
    entry_points={
        "console_scripts": [
            "microspec-api-example=microspeclib.examples.microspec_api:main",
            "microspec-lowlevel-api-example=microspeclib.examples.microspec_lowlevel_api:main",
            "microspec-cmdline=microspeclib.scripts.microspec_cmdline:main",
        ],
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={
        'Chromation': 'https://www.chromation.com/',
        'Thanks To': 'http://eruciform.com/',
    },
    license='MIT', # field in *.egg-info/PKG-INFO
    platforms=['Windows', 'Mac', 'Linux'], # legacy field in *.egg-info/PKG-INFO
)




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest, copy
from microspeclib.internal.payload import MicroSpecPayloadClassFactory, MicroSpecPayload, MicroSpecRepeatPayload

class MicroSpecTestPayloadFactory(unittest.TestCase):

  def test_payloadFactoryNormal(self):
    klass = MicroSpecPayloadClassFactory("command", 99, "grok", ["command_id","foo","bar","baz"], [1,1,2,4])
    assert klass.command_id == 99
    assert klass.__name__   == "grok"
    assert klass.variables  == ["command_id","foo","bar","baz"]
    assert klass.sizes      == [1,1,2,4]

  def test_payloadFactoryRepeat(self):
    klass = MicroSpecPayloadClassFactory("command", 99, "grok", ["command_id","foo","bar","baz"], [1,1,2,4], repeat={"foo":"bar"})
    assert klass.command_id == 99
    assert klass.__name__   == "grok"
    assert klass.variables  == ["command_id","foo","bar","baz"]
    assert klass.sizes      == [1,1,2,4]
    assert klass.repeat     == {"foo":"bar"}

class MicroSpecTestPayload(unittest.TestCase):

  def __init__(self, *args, **kwargs):
    super(MicroSpecTestPayload, self).__init__(*args, **kwargs)
    self.klass = MicroSpecPayloadClassFactory("command", 99, "grok", ["command_id","foo","bar","baz"], [1,1,2,4])

  def test_initEmpty(self):
    obj = self.klass()
    assert obj.foo        == None
    assert obj.bar        == None
    assert obj.baz        == None
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]

  def test_initKwargs(self):
    obj = self.klass(foo=1,bar=2,baz="0xFF")
    assert obj.foo        == 1
    assert obj.bar        == 2
    assert obj.baz        == 255
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]

  def test_initPayload(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    assert obj.foo        == 1
    assert obj.bar        == 2
    assert obj.baz        == 255
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]

  def test_unpack(self):
    obj = self.klass()
    obj.unpack(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    assert obj.foo        == 1
    assert obj.bar        == 2
    assert obj.baz        == 255
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]

  def test_iter(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    var = []
    for v in obj:
      var.append(v)
    assert var == ["command_id","foo","bar","baz"]

  def test_repr(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    s   = """<grok name=grok command_id=99 variables=['command_id', 'foo', 'bar', 'baz'] values={'command_id': 99, 'foo': 1, 'bar': 2, 'baz': 255} sizes=[1, 1, 2, 4] packformat=>BBHL length=8 packed=b'c\\x01\\x00\\x02\\x00\\x00\\x00\\xff'>"""
    assert repr(obj) == s

  def test_str(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    s   = """grok(command_id=99, foo=1, bar=2, baz=255)"""
    assert str(obj) == s

  def test_bytes(self):
    b = b'\x63\x01\x00\x02\x00\x00\x00\xFF'
    obj = self.klass(b)
    assert bytes(obj) == b

  def test_len(self):
    b = b'\x63\x01\x00\x02\x00\x00\x00\xFF'
    obj = self.klass(b)
    assert len(bytes(obj)) == len(b)

  def test_packformat(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    assert obj.packformat() == ">BBHL"

  def test_packvalues(self):
    obj = self.klass(b'\x63\x01\x00\x02\x00\x00\x00\xFF')
    assert obj.packvalues() == [99, 1, 2, 255]

  def test_pack(self):
    b = b'\x63\x01\x00\x02\x00\x00\x00\xFF'
    obj = self.klass(b)
    assert obj.pack() == b

  def test_equals(self):
    b = b'\x63\x01\x00\x02\x00\x00\x00\xFF'
    obj = self.klass(b)
    ob2 = self.klass(b)
    assert obj == ob2

    ob2.name = "fii"
    assert obj == ob2

    ob2 = self.klass(b)
    ob2.varsize[0] = 999
    assert obj != ob2

    ob2 = self.klass(b)
    ob2.command_id = 999
    assert obj == ob2

    ob2 = self.klass(b)
    ob2.variables = ["a","b","c","d"]
    assert obj == ob2

    ob2 = self.klass(b)
    ob2.sizes = [999,999,999,999]
    assert obj == ob2

  def test_packValues(self):
    obj = self.klass(foo=1, bar=2, baz=255)
    assert bytes(obj) == b'\x63\x01\x00\x02\x00\x00\x00\xFF'

  def test_packValuesMissing(self):
    obj = self.klass(foo=1, bar=2)
    assert bytes(obj) == b''

    obj = self.klass(foo=1, baz=255)
    assert bytes(obj) == b''

    obj = self.klass(bar=2, baz=255)
    assert bytes(obj) == b''

class MicroSpecTestRepeatPayload(MicroSpecTestPayload):

  def __init__(self, *args, **kwargs):
    super(MicroSpecTestRepeatPayload, self).__init__(*args, **kwargs)
    self.klass = MicroSpecPayloadClassFactory("command", 99, "grok", ["command_id","foo","bar","baz"], [1,1,2,4], repeat={"baz": "foo"})

  def test_initEmpty(self):
    obj = self.klass()
    assert obj.foo        == None
    assert obj.bar        == None
    assert obj.baz        == []
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]
    assert obj.repeat     == {"baz": "foo"}

  def test_initKwargs(self):
    obj = self.klass(foo=3,bar=2,baz=["0xFF",0,17])
    assert obj.foo        == 3
    assert obj.bar        == 2
    assert obj.baz        == [255,0,17]
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]
    assert obj.repeat     == {"baz": "foo"}

  def test_initPayload(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    assert obj.foo        == 3
    assert obj.bar        == 2
    assert obj.baz        == [255,0,17]
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]
    assert obj.repeat     == {"baz": "foo"}

  def test_unpack(self):
    obj = self.klass()
    obj.unpack(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    assert obj.foo        == 3
    assert obj.bar        == 2
    assert obj.baz        == [255,0,17]
    assert obj.command_id == 99
    assert obj.variables  == ["command_id","foo","bar","baz"]
    assert obj.sizes      == [1,1,2,4]

  def test_repr(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    s   = """<grok name=grok command_id=99 variables=['command_id', 'foo', 'bar', 'baz'] values={'command_id': 99, 'foo': 3, 'bar': 2, 'baz': [255, 0, 17]} sizes=[1, 1, 2, 4] packformat=>BBHLLL length=8 packed=b'c\\x03\\x00\\x02\\x00\\x00\\x00\\xff\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x11'>"""
    assert repr(obj) == s

  def test_str(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    s   = """grok(command_id=99, foo=3, bar=2, baz=[255, 0, 17])"""
    assert str(obj) == s

  def test_packformat(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    assert obj.packformat() == ">BBHLLL"

  def test_tooFewData(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    obj.foo = 4
    # NOTE: the "repeat" number doesn't follow the list
    #       and if the number is too large for the data, it will throw an exception
    assert obj.foo        == 4
    assert obj.baz        == [255,0,17]
    with pytest.raises(Exception) as excinfo:
      b = bytes(obj)
    assert "pack expected 7 items for packing (got 6)" in str(excinfo.value)

  def test_tooManyData(self):
    obj = self.klass(b'\x63\x03\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11')
    obj.foo = 4
    obj.baz = [255,0,17,42,99]
    # NOTE: the "repeat" number doesn't follow the list
    #       though the list does get chopped at bytes() or pack() time
    assert obj.foo        == 4
    assert obj.baz        == [255,0,17,42,99]
    b = bytes(obj)
    assert b == b'\x63\x04\x00\x02\x00\x00\x00\xFF\x00\x00\x00\x00\x00\x00\x00\x11\x00\x00\x00\x2A'

  def test_equals(self):
    super().test_equals()
    b = b'\x63\x01\x00\x02\x00\x00\x00\xFF'
    obj = self.klass(b)
    ob2 = self.klass(b)
    ob2.repeat = "xxx"
    assert obj == ob2

  def test_packValues(self):
    obj = self.klass(foo=1, bar=2, baz=[255])
    assert bytes(obj) == b'\x63\x01\x00\x02\x00\x00\x00\xFF'

  def test_packValuesMissing(self):
    obj = self.klass(foo=1, bar=2)
    with pytest.raises(Exception) as excinfo:
      b = bytes(obj)
    assert "pack expected 4 items for packing (got 3)" in str(excinfo.value)

    obj = self.klass(foo=1, baz=[255])
    assert bytes(obj) == b''

    obj = self.klass(bar=2, baz=[])
    assert bytes(obj) == b''














# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
# I commented these out because they look like the wrong path.
# sys.path.insert(0, os.path.abspath('../src'))
# sys.path.insert(1, os.path.abspath('../'))

# Adding these causes RTD built to fail with AttributeError: 'Sphinx' object has
# no attribute 'warn'
# sys.path.insert(0, os.path.abspath('../../tests/'))
# sys.path.insert(0, os.path.abspath('../../'))
# These also do not help.
# sys.path.insert(0, os.path.abspath('../../src/'))
# sys.path.insert(0, os.path.abspath('../../bin/'))
#
# Trying to avoid "PYTHONPATH=../tests:../" before "make clean html"


# -- Project information -----------------------------------------------------

project = 'microspeclib'
copyright = '2020, Chromation Inc'
author = 'Sean Cusack'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
"sphinx.ext.autodoc",
"sphinx.ext.coverage",
"sphinx.ext.napoleon",
# 'm2r', incompatible with Sphinx
#        see https://github.com/sphinx-doc/sphinx/issues/7420
"recommonmark", # replaces m2r
"sphinxcontrib.argdoc",
"sphinx_rtd_theme", # use the readthedocs theme
]

# Override readthedocs default: master_doc='contents.rst'
master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_logo = '_static/CHROMATION.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static/']

# Autodoc options
autodoc_default_options = {
    'no-inherited-members': True,
    'undoc-members': False,
}

def skip_payload_attributes(app, what, name, obj, skip, options):
  if what == "class" and getattr(obj, "__qualname__", None) is None:
    # Skip attributes, this includes internal things like __doc__ but
    # also the class non-function attributes like led_num, because Sphinx
    # renders them horribly, so we might as well just leave it in the
    # class docstring instead
    #
    # NOTE: for some reason, even though led_num is UNDOCUMENTED, leaving
    # out undoc-members still insists on including them
    return True

def setup(app):
  # NOTE: This is a bit hacky, but sphinxcontrib-argdoc depends on a deprecated
  # call in Sphinx, so this is a workaround
  from sphinx.util import logging
  logger = logging.getLogger("sphinxcontrib.argdoc")
  app.debug2 = logger.debug

  # Connect to the class attribute skipping function
  app.connect('autodoc-skip-member', skip_payload_attributes)

napoleon_numpy_docstring = True

navigation_depth = 1





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest, psutil, time, sys
from io import BytesIO
from microspeclib.internal.stream   import MicroSpecEmulatedStream, \
                                            MicroSpecSerialIOStream
from microspeclib.internal.emulator import MicroSpecEmulator
from microspeclib.datatypes         import *
from microspeclib.datatypes.command import CHROMASPEC_COMMAND_ID

#from microspeclib.internal.logger import CHROMASPEC_LOGGER_INTERNAL
#import logging

from test_bytesio_stream import MicroSpecTestBytesIOStream

@pytest.mark.skipif(sys.platform not in ["darwin","linux"], reason="Emulation currently only runs on linux and MacOS")
class MicroSpecTestEmulatedStream(MicroSpecTestBytesIOStream):
  def __init__(self, *args, **kwargs):
    # Note: 0.1 second is way more than needed, but most of these are tests of proper functionality,
    # not failures and partial reads. For those cases, the timeout is set explicitly.
    #
    # Note: for testing low-level reads and writes, write to the hardware directly here and read
    # from software, as the separate running process is reading from hardware and will respond to
    # writes to the software side, thus messing up the tests. However, if you pretend to be the
    # hardware side and only write, it's not reading from that side of the socket.
    # 
    # Note: this is testing the STREAM, not the EMULATOR, so separate test files will fully end-to-end
    # send a command and check the response.
    super().__init__(*args, **kwargs)
    self.num_socats_before = self.num_running_socats()
    self.emulator = MicroSpecEmulator()
    self.hardware = MicroSpecEmulatedStream(timeout=0.1, socat=True, fork=False)
    self.software = MicroSpecSerialIOStream(device=self.hardware.software, timeout=0.1)

  def num_running_socats(self):
    num = 0
    for proc in psutil.process_iter():
      try:
        if "socat" in proc.name().lower():
          num += 1
      except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
        pass
    return num

  def test_socatRunning(self):
    assert self.num_running_socats() > self.num_socats_before

  def test_emulatorPath(self):
    assert os.path.exists(self.hardware.hardware) == True
    assert os.path.exists(self.hardware.software) == True

  def test_partialReadBridgeNonzerostatusAndSensorReply(self):
    super().test_partialReadBridgeNonzerostatusAndSensorReply()
    # This (necessarily) leaves the buffer in a partial state, needs to be cleaned up
    self.hardware.stream.reset_input_buffer()
    self.hardware.buffer = b''
    self.hardware.current_command = []
    self.software.stream.reset_input_buffer()
    self.software.buffer = b''
    self.software.current_command = []

  # Note: until end of file, overrides of the BytesIO tests, so that we can use the same
  # tests without having to cut-and-paste:

  def generate_streams(self):
    return self.hardware, self.software

  def seek(self, stream, pos):
    return None

  def write_underlying(self, b, s, *args, **kwargs):
    return b.write(*args, **kwargs)

  def write_direct(self, b, s, *args, **kwargs):
    return b.write(*args, **kwargs)

  def sendCommand(self, b, s, *args, **kwargs):
    return b.sendCommand(*args, **kwargs)

  def sendReply(self, b, s, *args, **kwargs):
    return b.sendCommand(*args, **kwargs)

  def small_wait(self):
    time.sleep(0.1)

  def assert_getvalue(self, stream, value):
    pass

  def assert_readpos(self, stream, pos):
    pass
  
  def test_parameterStream(self):
    pass





# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.logger import CHROMASPEC_LOGGER_JSON as log
from .payload    import MicroSpecPayloadClassFactory as cfactory
from .util       import dehex
from .configutil import findConfig
import json

def globalizeJsonFile(filename):
  log.info("filename=%s", filename)
  with open(findConfig(filename)) as f:
    j = json.load(f)["globals"]
    log.debug("raw json=%s", j)
    jobjects = dict([k,dehex(v)] for k,v in j.items())
    log.info("return %s", jobjects)
    return jobjects
  
def enclassJsonFile(filename, protocol="command"):
  log.info("filename=%s protocol=%s", filename, protocol)
  with open(findConfig(filename)) as f:
    j      = json.load(f)
    p      = j["protocol"][protocol]
    byID   = {}
    byName = {}
    log.debug("raw json=%s", j)
    log.debug("raw json protocol section=%s", p)
    for k, v in p.items():
      log.debug("k=%s v=%s", k, v)
      c = cfactory(protocol, int(k), protocol.capitalize()+p[k]["name"], 
                   p[k]["variables"], p[k]["sizes"], p[k].get("repeat",None))
      byID[c.command_id] = c
      byName[c.name] = c
  log.info("return %s, %s", byID, byName)
  return byID, byName




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

import unittest, os, pytest, time
from timeit import default_timer as timer
from tabulate import tabulate
from microspeclib.expert            import MicroSpecExpertInterface
from microspeclib.internal.emulator import MicroSpecEmulator
from microspeclib.datatypes         import *
from microspeclib.datatypes.command import CHROMASPEC_COMMAND_ID

@pytest.mark.usefixtures("class_results")
class MicroSpecTestExpertInterface(unittest.TestCase):
  __test__ = False # Abstract test class #

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  @classmethod
  def setUpClass(cls):
    cls.setup = False
    cls.emulator = MicroSpecEmulator()
    cls.min      = 0

  def setUp(self):
    if self.__class__.setup:
      return
    command = CommandGetBridgeLED(led_num=255)
    for i in range(0,100):
      t1 = timer()    
      self.software.sendCommand(command)
      r  = self.software.receiveReply()
      t2 = timer()    
      r.led_setting = 0
      assert r == BridgeGetBridgeLED(status=1, led_setting=0)
      self.__class__.min += t2 - t1
    self.__class__.min /= 100
    self.results.append(["Expert."+command.__class__.__name__ + "(Reference)", self.min*1000])
    self.__class__.setup = True

  @classmethod
  def tearDownClass(self):
    print()
    print(tabulate(self.results, headers='firstrow', tablefmt="pipe"))

def generateTest(command_class):
  def test(self):
    command = command_class()
    if command_class is CommandSetSensorConfig:
      command.binning    = True
      command.gain       = Gain1x
      command.row_bitmap = 0x1F
    elif command_class is CommandSetExposure:
      command.cycles  = 1
    else:
      for var in command:
        command[var]  = 0
    replies = self.__class__.emulator.process(command)
    if replies:
      expected_reply = replies.pop()
    else:
      expected_reply = BridgeNull()
    avg = 0
    for i in range(0,100):
      t1 = timer()    
      self.software.sendCommand(command)
      r  = self.software.receiveReply()
      t2 = timer()    
      if command_class is CommandCaptureFrame:
        # Fake it - can't possibly predict capture data
        r.num_pixels = expected_reply.num_pixels
        r.pixels     = expected_reply.pixels
      assert r == expected_reply
      avg += t2 - t1
    avg /= 100
    avg -= self.__class__.min
    self.results.append(["Expert."+command.__class__.__name__, avg*1000])
  return test

for command_id, command_class in CHROMASPEC_COMMAND_ID.items():
  if command_id < 0:
    continue
  order = "0"
  if   command_class.__name__[0:12] == "CommandReset": order = "0"
  elif command_class.__name__[0:10] == "CommandSet":   order = "1"
  else:                                                order = "2"
  setattr(MicroSpecTestExpertInterface, "test_"+order+"sending"+command_class.__name__, generateTest(command_class))




#!/usr/bin/env python

'''Examples using the API.

USAGE
-----
After "pip install microspec", this script runs from the command
line with the command "microspec-api-example"

$ microspec-api-example
'''

def main():
    '''A few examples using the microspec API.

    Usage
    -----
    Execute this function at the command line with command
    "microspec-api-example":

    $ microspec-api-example

    Entry Point
    -----------
    The name of this function must match the attribute in
    `setup.py` `setuptools.setup()` argument `entry_points`.
    '''
    from microspeclib.simple    import MicroSpecSimpleInterface
    from microspeclib.datatypes import LEDOff, LEDGreen, LEDRed, Gain1x

    si = MicroSpecSimpleInterface(timeout=0.1)
    #si = MicroSpecSimpleInterface(timeout=0.1, device="COM3")
    #si = MicroSpecSimpleInterface(timeout=0.1, device="/dev/cu.usbserial-CHROMATION09310")
    #si = MicroSpecSimpleInterface(timeout=0.1, emulation=True)

    print(si.setBridgeLED(led_num=0, led_setting=LEDOff))
    print(si.setSensorLED(led_num=0, led_setting=LEDGreen))
    print(si.setSensorLED(led_num=1, led_setting=LEDRed))
    print(si.getBridgeLED(led_num=0))
    print(si.getSensorLED(led_num=0))
    reply = si.getSensorLED(led_num=1)
    print(reply.status)
    print(reply.led_setting)

    print(si.setSensorConfig(binning=True, gain=Gain1x, row_bitmap=0x1F))
    print(si.setExposure(cycles=100))

    import time
    for i in range(0,5):
        reply = si.captureFrame()
        print(reply.status)
        print(reply.num_pixels)
        pixels = [str(pixel) for pixel in reply.pixels]
        print("first 3 pixels: %s"%(",".join(pixels[:3])))
        time.sleep(0.5)

if __name__ == '__main__':
    main()




# Copyright 2020 by Chromation, Inc
# All Rights Reserved by Chromation, Inc

from microspeclib.datatypes  import *
from microspeclib.logger     import CHROMASPEC_LOGGER_STREAM as log
from microspeclib.exceptions import MicroSpecConnectionException, MicroSpecEmulationException
from struct       import pack, unpack
from io           import BytesIO
from serial       import Serial
from serial.tools import list_ports

import subprocess
import tempfile
import atexit
import select
import sys 
import os

class MicroSpecStream(object):
  def __init__(self, stream):
    log.info("stream=%s", stream)
    self.stream = stream
    self.buffer = b''
    log.info("return")

  def read(self, bytelen=0, *args, **kwargs):
    log.info("bytelen=%s args=%s kwargs=%s", bytelen, args, kwargs)
    if bytelen:
      if bytelen > len(self.buffer):
        self.buffer += self.stream.read(bytelen-len(self.buffer), *args, **kwargs)
      log.info("return buffer[:%s]=%s",bytelen,self.buffer[:bytelen])
      return self.buffer[:bytelen]
    else:
      self.buffer += self.stream.read(*args, **kwargs)
      log.info("return buffer=%s",self.buffer)
      return self.buffer

  def consume(self, bytelen):
    log.info("bytelen=%s", bytelen)
    self.buffer = self.buffer[bytelen:]
    log.info("return")

  def pushback(self, buf):
    log.info("buf=%s", buf)
    self.buffer = buf + self.buffer
    log.info("return")

  def write(self, buf, *args, **kwargs):
    log.info("buf=%s args=%s kwargs=%s", buf, args, kwargs)
    buf = self.stream.write(bytes(buf), *args, **kwargs)
    log.info("return buf=%s", buf)
    return buf

  def receiveCommand(self):
    log.info("read 1")
    command_id = self.read(1)
    if len(command_id) < 1:
      log.info("Did not read one byte of command_id")
      return None
    klass = getCommandByID(unpack(b'B', command_id)[0])
    if not klass:
      log.error("Command ID not recognized: %s", command_id)
      return None
    command = klass()
    self.consume(1)
    if len(command) > 1:
      buf = command_id + self.read(len(command) - 1)
    else:
      buf = command_id
    if len(buf) < len(command):
      log.info("Read only %s of %s bytes", len(buf), len(command))
      self.pushback(command_id)
      return None
    try:
      log.info("unpack %s", buf)
      command.unpack(buf)
    except Exception as e:
      log.info("Cannot unpack buf=%s exception=%s", buf, str(e))
      return None
    self.consume(len(command) - 1)
    log.info("return %s", command)
    return command

  def sendCommand(self, command):
    log.info("command=%s", command)
    return self.write(command)

  def sendReply(self, reply):
    log.info("reply=%s", reply)
    result = self.write(reply)
    log.info("return result=%s", result)
    return result

  def receiveReply(self, command_id):
    log.info("command_id=%s", command_id)
    bridge_klass = getBridgeReplyByID(command_id)
    sensor_klass = getSensorReplyByID(command_id)
    if not bridge_klass:
      log.error("Command ID not recognized: %s", command_id)
      return None
    if bridge_klass is BridgeNull:
      log.info("bridge reply is BridgeNull")
      return bridge_klass()
    bridgebuf = self.read(0)
    log.info("bridgebuf=%s", bridgebuf)
    try:
      bridge_reply = bridge_klass(bridgebuf)
    except Exception as e:
      log.info("bridgebuf=%s exception=%s", bridgebuf, str(e))
      return None
    if not hasattr(bridge_reply, "status") or bridge_reply.status is None:
      log.info("bridge reply is empty")
      return None
    bridgelen = len(bytes(bridge_reply))
    self.consume(bridgelen)
    bridgebuf = bridgebuf[0:bridgelen]
    if bridge_reply.status != 0 or not sensor_klass:
      log.info("return bridge_reply=%s", bridge_reply)
      return bridge_reply
    sensorbuf = self.read(0)
    log.info("bridgebuf=%s", sensorbuf)
    try:
      sensor_reply = sensor_klass(sensorbuf)
    except Exception as e:
      log.info("sensorbuf=%s exception=%s pushing back bridgebuf=%s", 
        sensorbuf, str(e), bridgebuf)
      self.pushback(bridgebuf)
      return None
    if sensor_reply.status is None:
      log.info("sensor reply is empty")
      self.pushback(bridgebuf)
      return None
    self.consume(len(bytes(sensor_reply)))
    log.info("return sensor_reply=%s", sensor_reply)
    return sensor_reply

# Single-threaded test functionality, assumes no partially
# interleaved reading and writing by multiple threads
class MicroSpecBytesIOStream(MicroSpecStream):
  def __init__(self, stream=None):
    log.info("stream=%s", stream)
    if not stream:
      stream = BytesIO()
    self.readpos  = 0
    self.writepos = 0
    super().__init__(stream)

  def read(self, bytelen=0, *args, **kwargs):
    log.info("bytelen=%s args=%s kwargs=%s", bytelen, args, kwargs)
    log.info("readpos=%s", self.readpos)
    self.stream.seek(self.readpos)
    pre  = self.stream.tell()
    buf  = super().read(bytelen=bytelen, *args, **kwargs)
    post = self.stream.tell()
    self.readpos += post - pre
    log.info("return buf=%s", buf)
    return buf

  def write(self, buf, *args, **kwargs):
    log.info("buf=%s args=%s kwargs=%s", buf, args, kwargs)
    log.info("writepos=%s", self.writepos)
    self.stream.seek(self.writepos)
    result = super().write(bytes(buf), *args, **kwargs)
    self.writepos += result
    log.info("return result=%s", result)
    return result

class MicroSpecSerialIOStream(MicroSpecStream):
  def __init__(self, serial_number=None, device=None, timeout=0, *args, **kwargs):
    log.info("serial_number=%s device=%s timeout=%s args=%s kwargs=%s", serial_number, device, timeout, args, kwargs)
    self.serial          = Serial(*args, **kwargs)
    self.serial.baudrate = 115200
    self.serial.timeout  = timeout
    self.serial.port     = None
    if serial_number:
      ports = list(list_ports.grep(serial_number))
      if ports:
        self.serial.port = ports[0].device
        self.serial.serial_number = ports[0].serial_number
        log.info("search for serial_number=%s found port=%s", serial_number, self.serial.port)
    elif device:
      self.serial.port = device
      log.info("using device=%s", device)
    else:
      ports = list(list_ports.grep("CHROMATION"))
      if ports:
        self.serial.port = ports[0].device
        self.serial.serial_number = ports[0].serial_number
        log.info("defaulting to searching for CHROMATION hardware, "
                "found port=%s, serial_number=",
                self.serial.port, self.serial.serial_number
                )
      if not self.serial.port:
        for port in list_ports.comports(True):
          if port.vid == 1027 and port.pid == 24597:
            self.serial.port = port.device
            break
    if not self.serial.port:
      log.error("Cannot find CHROMATION device")
      raise MicroSpecConnectionException("Cannot find CHROMATION device")
    try:
      self.serial.open()
    except Exception as e:
      log.error("Cannot open device: check if port is already in use, such as another MicroSpec interface is using it: %s"%(e))
      raise MicroSpecConnectionException(str(e))
    super().__init__(self.serial)
    log.info("return")

  def read(self, bytelen=0, *args, **kwargs):
    log.info("bytelen=%s args=%s kwargs=%s", bytelen, args, kwargs)
    if not bytelen:
      waiting  = self.stream.inWaiting()
      inbuffer = len(self.buffer)
      waitfor  = waiting+inbuffer
      log.info("serial stream bytelen=%s waiting=%s inbuffer=%d setting to %s", bytelen, waiting, inbuffer, waitfor)
      bytelen = waitfor
    return super().read(bytelen, *args, **kwargs)

class MicroSpecEmulatedStream(MicroSpecSerialIOStream):
  def __init__(self, hardware=None, software=None, timeout=None, socat=False, fork=False, *args, **kwargs):
    log.info("hardware=%s software=%s timeout=%s args=%s kwargs=%s", hardware, software, timeout, args, kwargs)
    cleanup_hardware = False
    cleanup_software = False
    cleanup_tempdir  = False

    if not hardware or not software:
      tempdir = tempfile.mkdtemp()
      cleanup_tempdir = True

    if not hardware:
      hardware = os.path.join(tempdir, "chromation.hardware")
      cleanup_hardware = True

    if not software:
      software = os.path.join(tempdir, "chromation.software")
      cleanup_software = True

    if socat:
      log.info("forking socat process")
      #TODO: make this work for PC as well as MAC
      # Note: the -D is there to print something, anything, to stderr, so we can wait for it
      socat = subprocess.Popen(["socat", "-D", "PTY,raw,echo=0,link=%s"%(hardware),
                                               "PTY,raw,echo=0,link=%s"%(software)],
                               stderr=subprocess.PIPE)
  
      # Wait at most a second for socat to come up and announce itself, by which time the
      # files it needs to create should be there:
      r, w, x = select.select([socat.stderr],[],[],1)
      #import pdb; pdb.set_trace();
      if not r:
        raise MicroSpecEmulationException("Cannot create socat process to mediate USB emulation!")
  
      def cleanup():
        log.warning("Cleanup: killing socat process")
        socat.kill()
        if cleanup_hardware:
          try:
            log.warning("Cleanup: removing hardware file %s"%(hardware))
            os.remove(hardware)
          except:
            pass # possibly a race condition with socat dying and not removing this quickly enough
        if cleanup_software:
          try:
            log.warning("Cleanup: removing software file %s"%(software))
            os.remove(software)
          except:
            pass # possibly a race condition with socat dying and not removing this quickly enough
        if cleanup_tempdir:
          if os.path.isdir(tempdir):
            log.warning("Cleanup: removing temp directory %s"%(tempdir))
            os.rmdir(tempdir)
        log.warning("Cleanup: done")
      atexit.register(cleanup)
      self.socat = socat

    if fork:
      log.info("forking emulator process")
      # -p so that we can capture the stdout to make sure it's up and running
      fork = subprocess.Popen(["microspec_emulator.py", "-t", "10", "-p", "-f", hardware],
                              stdout=subprocess.PIPE)

      # Wait at most a second for emulator to come up and announce itself, by which time it
      # should have it's socket open and waiting
      r, w, x = select.select([fork.stdout],[],[],1)
      #import pdb; pdb.set_trace();
      if not r:
        raise MicroSpecEmulationException("Cannot create emulator process for USB emulation!")
  
      def cleanup():
        log.warning("Cleanup: killing emulator process")
        fork.kill()
        log.warning("Cleanup: done")
      atexit.register(cleanup)
      self.fork = fork

    self.hardware = hardware
    self.software = software

    super().__init__(device=hardware, timeout=timeout)
